/**
 * Copyright (c) 2006, 2009 Bob Swift
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright notice,
 *       this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *              notice, this list of conditions and the following disclaimer in the
 *            documentation and/or other materials provided with the distribution.
 *     * The names of contributors may not be used to endorse or promote products
 *           derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Created on March 21, 2008 by Bob Swift
 */

package org.swift.jira.cli;

import java.io.File;
import java.io.IOException;
import java.net.URLConnection;
import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import javax.activation.MimetypesFileTypeMap;
import javax.xml.rpc.ServiceException;

import org.swift.common.cli.CliClient;
import org.swift.common.cli.CliUtils;
import org.swift.common.cli.soap.AbstractSoapClient;

import com.dolby.jira.net.soap.jira.AbstractNamedRemoteEntity;
import com.dolby.jira.net.soap.jira.JiraSoapService;
import com.dolby.jira.net.soap.jira.JiraSoapServiceServiceLocator;
import com.dolby.jira.net.soap.jira.RemoteAttachment;
import com.dolby.jira.net.soap.jira.RemoteAuthenticationException;
import com.dolby.jira.net.soap.jira.RemoteComment;
import com.dolby.jira.net.soap.jira.RemoteComponent;
import com.dolby.jira.net.soap.jira.RemoteCustomFieldValue;
import com.dolby.jira.net.soap.jira.RemoteField;
import com.dolby.jira.net.soap.jira.RemoteFieldValue;
import com.dolby.jira.net.soap.jira.RemoteFilter;
import com.dolby.jira.net.soap.jira.RemoteGroup;
import com.dolby.jira.net.soap.jira.RemoteIssue;
import com.dolby.jira.net.soap.jira.RemoteIssueType;
import com.dolby.jira.net.soap.jira.RemoteNamedObject;
import com.dolby.jira.net.soap.jira.RemotePermissionException;
import com.dolby.jira.net.soap.jira.RemotePermissionScheme;
import com.dolby.jira.net.soap.jira.RemotePriority;
import com.dolby.jira.net.soap.jira.RemoteProject;
import com.dolby.jira.net.soap.jira.RemoteResolution;
import com.dolby.jira.net.soap.jira.RemoteScheme;
import com.dolby.jira.net.soap.jira.RemoteServerInfo;
import com.dolby.jira.net.soap.jira.RemoteStatus;
import com.dolby.jira.net.soap.jira.RemoteUser;
import com.dolby.jira.net.soap.jira.RemoteVersion;
import com.dolby.jira.net.soap.jira.RemoteWorklog;

/**
 * Soap Client - parseArguments function for parameters, defaults, and documentation - usage java -jar release/jira-cli-x.x.x.jar --help - see the README.txt
 * file for more information - to reuse login token jira --action login > token jira --action xxxxx -x < token TODO: File propertyFile = new File(...)
 * PropertyDefaultSource pds = new PropertyDefaultSource(propertyFile.getAbsolutePath(), true); jsap.registerDefaultSource(pds); jsap.parse ...
 * http://www.koders.com/java/fid78821B6E525479085FCF575197089FF8C0B36123.aspx
 */
public class JiraClient extends AbstractSoapClient {

    protected JiraSoapService service;
    protected RemoteUser remoteUser = null;
    protected JiraRestClient guest = null;
    protected JiraUserHelper userHelper = null;
    protected JiraPluginHelper pluginHelper = null;

    /**
     * Guest requests - TODO - not needed yet, always do normal setup even for guest request
     */
    protected ArrayList<Action> guestRequestList = new ArrayList<Action>() {
        private static final long serialVersionUID = 1L;
        {
            // add(Action.ADDCOMPONENT);
        }
    };

    /**
     * Command entry point - use as model for subclass implementation
     * 
     * @param args - command line arguments
     */
    public static void main(String[] args) {
        if ((args == null) || (args.length == 0)) {
            args = new String[] {"--help"};
        }

        ExitCode code = new JiraClient().doWork(args);

        System.exit(code.value());
    }

    /**
     * Get a new instance of this client. Used for run support.
     * 
     * @return new client instance
     */
    @Override
    public CliClient getNewClient() {
        return new JiraClient();
    }

    /**
     * Setup for request processing and then handle the request - this should be overridden by subclass for specific handling
     * 
     * @return - appropriate exit code based on processing
     */
    /*
     * @Override /*protected ExitCode process() { try { // if (guestRequest()) { // setupGuest(); // } else { setup(); // } } catch (Exception exception) {
     * err.println("An error occurred during setup.\nCause: " + exception.toString()); return ExitCode.CLIENT_EXCEPTION; } try { out.println("Token: " + token);
     * String message = handleRequest(); out.println(message); } catch (Exception exception) { if (verbose) { exception.printStackTrace(); }
     * err.println("An error occurred handling request.\nCause: " + exception.toString()); return ExitCode.CLIENT_EXCEPTION; } return ExitCode.SUCCESS; }
     */

    // TODO not needed ?
    protected boolean guestRequest() {
        boolean result;
        result = guestRequestList.contains(Action.toAction(getString("action").toUpperCase()));
        return result;
    }

    /**
     * Get the guest client, create if it doesn't exist. It is initialized with the current jsap, so has access to all parameters
     * 
     * @return rest client
     */
    protected JiraRestClient getGuest() {
        if (guest == null) {
            guest = new JiraRestClient(jsap, jsapResult);
        }
        return guest;
    }

    /**
     * Get the guest client, create if it doesn't exist. It is initialized with the current jsap, so has access to all parameters
     * 
     * @return rest client
     */
    protected JiraUserHelper getUserHelper() {
        if (userHelper == null) {
            userHelper = new JiraUserHelper(this, verbose, service, token);
        }
        return userHelper;
    }

    /**
     * Get plugin helper, create if it doesn't exist.
     * 
     * @return plugin helper
     * @throws java.rmi.RemoteException
     * @throws RemoteException
     * @throws InvalidSessionException
     */
    protected JiraPluginHelper getPluginHelper() {
        if (pluginHelper == null) {
            pluginHelper = new JiraPluginHelper(getGuest());
        }
        return pluginHelper;
    }

    /**
     * Get the name of the client that will appear in help text - command name
     * 
     * @return client name
     */
    @Override
    protected String getClientName() {
        return "jira";
    }

    /**
     * Extension to server address that is the location of the service to use - this is the default location that the user can override by parameter
     * 
     * @return - service url extention - example: /api/rest/
     */
    @Override
    protected String getDefaultServiceExtension() {
        return "/rpc/soap/jirasoapservice-v2"; // JIRA soap service
    }

    /**
     * Setup for remote service - this must be overridden by subclass - Example for Confluence: ConfluenceSoapServiceServiceLocator serviceLocator = new
     * ConfluenceSoapServiceServiceLocator(); serviceLocator.setConfluenceserviceV1EndpointAddress(address); service = serviceLocator.getConfluenceserviceV1();
     */
    @Override
    protected void serviceSetup(String address) throws ClientException, RemoteException {

        try {
            JiraSoapServiceServiceLocator serviceLocator = new JiraSoapServiceServiceLocator();
            serviceLocator.setJirasoapserviceV2EndpointAddress(address);
            service = serviceLocator.getJirasoapserviceV2();
        } catch (ServiceException exception) {
            throw new RemoteServiceException(exception.getMessage());
        }
    }

    /**
     * Login to the remote service - this must be overridden by subclass - Example for soap: String user = getString("user"); String password =
     * getString("password"); token = service.login(user, password); if (verbose) { out.println("Successful login to: " + address + " by user: " + user); }
     */
    @Override
    protected void serviceLogin() throws ClientException, RemoteException {

        String user = getString("user");
        String password = getString("password");
        token = service.login(user, password);
        if (verbose) {
            out.println("Successful login to: " + address + " by user: " + user);
        }
    }

    /**
     * Logout of remote service - this must be overridden by subclass - Example for soap: service.logout(token);
     */
    @Override
    protected void serviceLogout() throws RemoteException {
        service.logout(token);
    }

    /**
     * Enumerate all the valid actions - all command actions must be listed in this enumeration - this needs to be overridden in a subclass - borrow switch
     * technique from http://www.xefer.com/2006/12/switchonstring
     */
    protected enum Action {
        LOGIN, LOGOUT, RUN, RUNFROMSQL,

        GETSERVERINFO,

        // Projects
        CREATEPROJECT, DELETEPROJECT, GETPROJECTLIST, ADDVERSION, ADDCOMPONENT, DELETECOMPONENT, COPYCOMPONENTS,

        // Issues
        CREATEISSUE, CLONEISSUE, UPDATEISSUE, PROGRESSISSUE, ADDCOMMENT, ADDATTACHMENT, GETISSUE, GETFIELDVALUE, SETFIELDVALUE, GETAVAILABLESTEPS,

        // Work log
        ADDWORK,

        // Links
        LINKISSUE, DELETELINK,

        // Filters
        GETISSUELIST,

        // Comments
        GETCOMMENTLIST, GETCOMMENTS,

        // Attachments
        GETATTACHMENT, GETATTACHMENTLIST,

        // user management
        ADDUSER, ADDUSERWITHFILE, REMOVEUSER, REMOVEUSERWITHFILE, ADDGROUP, REMOVEGROUP, ADDUSERTOGROUP, ADDUSERTOGROUPWITHFILE, REMOVEUSERFROMGROUP, REMOVEUSERFROMGROUPWITHFILE,

        GETUSERLIST, LISTUSERS, // LISTUSERS is depreciated

        // Plugins - experimental
        GETPLUGINLIST, GETPLUGINDOWNLOAD,

        NOTFOUND; // the last one

        private static Action toAction(String string) {
            try {
                return valueOf(string);
            } catch (Exception exception) {
                return NOTFOUND;
            }
        }
    }

    /**
     * Map the user request to implementing code - this should be overridden by subclass for specific handling
     * 
     * @return - the output from the request that will be put on the out stream
     */
    @Override
    protected String handleRequest() throws ClientException, RemoteException, RemoteRestException {

        String action = getString("action").toUpperCase();

        switch (Action.toAction(action)) {

        case LOGIN:
            return login();
        case LOGOUT:
            return logout();
        case RUN:
            return run(jsapResult.userSpecified("continue"), jsapResult.userSpecified("quiet"), jsapResult.userSpecified("simulate"));
        case RUNFROMSQL:
            // file is required if sql is not provided, this is handled in the function
            return runFromSql(getString("sql"), getRequiredString("driver"), getString("url"), getString("host"), getString("port"),
                    getRequiredString("database"), getString("user"), getString("password"), jsapResult.userSpecified("continue"),
                    jsapResult.userSpecified("quiet"), jsapResult.userSpecified("simulate"));

        case GETSERVERINFO:
            return getServerInfo();

        case CREATEPROJECT:
            return createProject();
        case DELETEPROJECT:
            return deleteProject();
        case GETPROJECTLIST:
            return getProjectList();
        case ADDVERSION:
            return addVersion();
        case ADDCOMPONENT:
            return addComponent(getRemoteProject(getRequiredString("project")), getRequiredString("name"), getString("description"), getString("lead"));
        case DELETECOMPONENT:
            return deleteComponent();
        case COPYCOMPONENTS:
            return copyComponents();

        case CREATEISSUE:
            return createIssue();
        case CLONEISSUE:
            return cloneIssue();
        case UPDATEISSUE:
            return updateIssue();
        case PROGRESSISSUE:
            return progressIssueInWorkflow();
        case ADDCOMMENT:
            return addComment();
        case ADDATTACHMENT:
            return addAttachment();
        case GETISSUE:
            return getIssue();
        case GETFIELDVALUE:
            return getFieldValue();
        case SETFIELDVALUE:
            return setFieldValue();
        case GETAVAILABLESTEPS:
            return getAvailableSteps();
        case GETCOMMENTLIST:
            return getCommentList();
        case GETCOMMENTS:
            return getComments();
        case GETATTACHMENT:
            return getAttachment();
        case GETATTACHMENTLIST:
            return getAttachmentList();

        case LINKISSUE:
            return linkIssue();
        case DELETELINK:
            return deleteLink();

        case GETISSUELIST:
            return getIssueList();

        case ADDWORK:
            return addWork();

            // User and group management
        case ADDUSER:
            return getUserHelper().addUser(getRequiredString("userId"), getString("userPassword"), getRequiredString("userEmail"), getString("userFullName"));
        case ADDUSERWITHFILE:
            return getUserHelper().addUserWithFile(getRequiredString("file"), jsapResult.userSpecified("autoGroup"));
        case REMOVEUSER:
            return getUserHelper().removeUser(getRequiredString("userId"));
        case REMOVEUSERWITHFILE:
            return getUserHelper().removeUserWithFile(getRequiredString("file"));
        case ADDGROUP:
            return getUserHelper().addGroup(getRequiredString("group"));
        case REMOVEGROUP:
            return getUserHelper().removeGroup(getRequiredString("group"), getString("defaultGroup"));
        case ADDUSERTOGROUP:
            return getUserHelper().addUserToGroup(getRequiredString("userId"), getRequiredString("group"), jsapResult.userSpecified("autoGroup"));
        case ADDUSERTOGROUPWITHFILE:
            return getUserHelper().addOrRemoveUserToOrFromGroupWithFile(getRequiredString("file"), true, jsapResult.userSpecified("autoGroup"));
        case REMOVEUSERFROMGROUP:
            return getUserHelper().removeUserFromGroup(getRequiredString("userId"), getRequiredString("group"));
        case REMOVEUSERFROMGROUPWITHFILE:
            try {
                return getUserHelper().addOrRemoveUserToOrFromGroupWithFile(getRequiredString("file"), false, jsapResult.userSpecified("autoGroup"));
            } catch (NullPointerException e) {
                e.printStackTrace();
            }

        case GETUSERLIST:
            return getUserList();
        case LISTUSERS:
            return getUserList(); // depreciated

        case GETPLUGINLIST:
            return getPluginHelper().getPluginList(getString("plugin"), jsapResult.getInt("outputFormat"), jsapResult.getInt("count"));

        case GETPLUGINDOWNLOAD:
            return getPluginHelper().getPluginDownload(getString("plugin"), getString("version"), getString("file"));

        case NOTFOUND:
        default:
            throw new InvalidActionClientException("Invalid action specified: " + action + ".  Use --help for more information.");
        }
    }

    /**
     * Map the action to the action help text - this should be overridden by subclass for specific handling
     * 
     * @action - action name
     * @return - the formatted help text for the action
     */
    protected String getActionHelp(Action action) {

        String standardIssueFields = "priority, reporter, assignee, description, components, affectsVersions, fixVersions, environment,\n\t date, dateFormat, custom, autoVersion, autoComponent";
        String standardIssueFieldsUpdateOnly = "type, summary, resolution,\n\t ";
        String fileAndReplace = ", findReplace, file, encoding";

        switch (action) {
        // name, help, required parameters, optional parameters
        case LOGIN:
            return formatActionHelp("login", "Login to remote server. Returns login token.", "password", "user");
        case LOGOUT:
            return formatActionHelp("logout", "Logout of remote server.", "", "");
        case RUN:
            return formatActionHelp("run", "Run script.", "file", "continue, quiet, simulate, encoding");
        case RUNFROMSQL:
            return formatActionHelp("runFromSql", "Run script generated from SQL.", "sql or file, url or host, driver, database",
                    "host, port, url, continue, quiet, simulate, encoding");

        case GETSERVERINFO:
            return formatActionHelp("getServerInfo", "Get information about the JIRA server.", "", "");

        case CREATEPROJECT:
            return formatActionHelp("createProject", "Create a new project with key provided by project parameter.", "project, lead",
                    "name, permissionScheme, notificationScheme, issueSecurityScheme");
        case DELETEPROJECT:
            return formatActionHelp("deleteProject", "Delete a project.", "project", "");
        case GETPROJECTLIST:
            return formatActionHelp("getProjectList", "List defined projects.", "", "file, encoding");
        case ADDVERSION:
            return formatActionHelp("addVersion", "Add a new version to a project.", "project, name", "after, date, dateFormat");
        case ADDCOMPONENT:
            return formatActionHelp("addComponent", "Add component to a project.", "project, name", "description, lead");
        case DELETECOMPONENT:
            return formatActionHelp("deleteComponent", "Delete component from a project.", "project, name", "");
        case COPYCOMPONENTS:
            return formatActionHelp("copyComponents", "Copy all or some components from one project to another.", "project, toProject", "components");

        case CREATEISSUE:
            return formatActionHelp("createIssue", "Create a new issue for a project.", "project, type, summary", standardIssueFields + fileAndReplace);
        case CLONEISSUE:
            return formatActionHelp("cloneIssue", "Create a new issue by copying an existing issue.", "issue", "copyLinks, copyAttachments");
        case UPDATEISSUE:
            return formatActionHelp("updateIssue", "Update an existing issue.", "issue", standardIssueFieldsUpdateOnly + standardIssueFields + fileAndReplace);
        case PROGRESSISSUE:
            return formatActionHelp("progressIssue", "Progress issue through workflow.", "issue, step", standardIssueFieldsUpdateOnly + standardIssueFields
                    + "comment" + fileAndReplace);
        case ADDCOMMENT:
            return formatActionHelp("addComment", "Add a comment to an issue.", "issue", "comment, role, group" + fileAndReplace);
        case ADDATTACHMENT:
            return formatActionHelp("addAttachment", "Add an attachment to an issue.", "issue, file", "findReplace, name, encoding");
        case GETISSUE:
            return formatActionHelp("getIssue", "Get information about an existing issue.", "issue", "file, dateFormat, asVersion, encoding");
        case GETFIELDVALUE:
            return formatActionHelp("getFieldValue", "Get field value for an issue.", "issue, field", "file, dateFormat, asVersion, encoding");
        case SETFIELDVALUE:
            return formatActionHelp("setFieldValue", "Set custom field value for an issue.", "issue, field, file or values", "asVersion, asComponent, encoding");
        case GETAVAILABLESTEPS:
            return formatActionHelp("getAvailableSteps", "Get available workflow steps for issue.", "issue", "file, encoding");
        case GETCOMMENTLIST:
            return formatActionHelp("getCommentList", "List of comment information (id, dates, ...) for an issue.", "issue", "file, dateFormat, encoding");
        case GETCOMMENTS:
            return formatActionHelp("getComments", "Get a formatted string of all comment text for an issue.", "issue", "file, dateFormat, encoding");
        case GETATTACHMENT:
            return formatActionHelp("getAttachment", "Get lastest attachment by name or id for an issue.", "issue, file", "name, encoding");
        case GETATTACHMENTLIST:
            return formatActionHelp("getAttachmentList", "List all attachments for an issue.", "issue, file", "dateFormat, encoding");

        case LINKISSUE:
            return formatActionHelp("linkIssue", "Link an issue to another issue.", "issue, toIssue, link", "comment");
        case DELETELINK:
            return formatActionHelp("deleteLink", "Remove link to another issue.", "issue, toIssue, link", "");

        case GETISSUELIST:
            return formatActionHelp("getIssueList", "List issues for a filter or a search.", "filter or search", "file, dateFormat, encoding");

        case ADDWORK:
            return formatActionHelp("addWork", "Add work log entry.", "issue, timeSpent", "comment, date, dateFormat");

            // User and group management
        case ADDUSER:
            return formatActionHelp("addUser", "Add a new user.", "userId", "userFullName, userEmail, userPassword");
        case ADDUSERWITHFILE:
            return formatActionHelp("addUserWithFile", "Add users from comma separated file.", "file", "encoding");
        case REMOVEUSER:
            return formatActionHelp("removeUser", "Remove a user.", "userId", "");
        case REMOVEUSERWITHFILE:
            return formatActionHelp("removeUserWithFile", "Remove users from comma separate file.", "file", "encoding");
        case ADDGROUP:
            return formatActionHelp("addGroup", "Add a new group.", "group", "");
        case REMOVEGROUP:
            return formatActionHelp("removeGroup", "Remove a group.", "group", "defaultGroup");
        case ADDUSERTOGROUP:
            return formatActionHelp("addUserToGroup", "Add user to a group.", "userId, group", "autoGroup");
        case ADDUSERTOGROUPWITHFILE:
            return formatActionHelp("addUserToGroupWithFile", "Add users to groups from comma separated file.", "file", "autoGroup, encoding");
        case REMOVEUSERFROMGROUP:
            return formatActionHelp("removeUserFromGroup", "Remove user from a group.", "userId, group", "");
        case REMOVEUSERFROMGROUPWITHFILE:
            return formatActionHelp("removeUserFromGroupWithFile", "Remove users from groups from comma separated file.", "file", "encoding");

        case GETUSERLIST:
            return formatActionHelp("getUserList", "List users in a group.", "group", "file, encoding");

        case GETPLUGINLIST:
            return formatActionHelp("getPluginList", "Get list of plugins. Use outputFormat=2 for extended information (longer time required).", "",
                    "plugin, file, outputFormat, count");

        case GETPLUGINDOWNLOAD:
            return formatActionHelp("getPluginDownload", "Get url to download the plugin version. Version defaults to the latest version.", "plugin", "version");

        case NOTFOUND:
        default:
            return "";
        }
    }

    /**
     * Append action help values to the string buffer - this is used so that the Action enum for this class is used
     * 
     * @param string buffer to append the help text
     * @return general help for this client
     */
    @Override
    protected void appendActionHelpValues(StringBuilder help) {
        for (Action action : Action.values()) {
            help.append(getActionHelp(action));
        }
    }

    /**
     * Add parameters to the parameterList - Options must have a value - Switches do NOT have a value - Note that the long version is expressed as: --xxx value,
     * --xxx "value", or --xxx and that short flag version is expressed as: -x value, -x "value", or -x
     */
    @Override
    protected void addParameters() {

        super.addParameters(); // add all the standard parameters for superclass client

        // name and help text
        addOption("project", "Project name, key, or id");
        addOption("toProject", "Project name, key, or id to copy to");
        addOption("name", "Name");
        addOption("description", "Description");
        // addOption("url", "URL"); TODO
        addOption("lead", "Project lead user id");
        addOption("after", "Version name or id to add a version after");

        addOption("issue", "Issue key or id");
        addOption("toIssue", "Link destination issue key or id");
        addOption("parent", "Parent issue key or id");
        addOption("summary", "Summary of issue");
        addOption("priority", "Issue priority - name or id");
        addOption("reporter", "Issue reporter user id");
        addOption("assignee", "Issue assignee user id");
        addOption("environment", "Issue environment");
        addOption("components", "Components - comma separated names or ids");
        addOption("affectsVersions", "Affects versions - comma separated names or ids");
        addOption("fixVersions", "Fix versions - comma separated names or ids");
        addOption("custom",
                "Custom fields - comma separated key:value pairs. Key can be field name or id. Single quote the key:value pair if it contains a comma (,) or line breaks.");
        addOption("field", "Field name or id. For some actions, this parameter must be a custom field.");
        addOption("date", "Date for version or due date for issue");
        addOption("dateFormat", "Format string for date in Java SimpleDateFormat, default is client date format");
        addOption("type", "Issue type - name or id");
        addOption("resolution", "Resolution name or id");
        addOption("step", "Workflow step - name or id");
        addOption("comment", "Comment for an issue");
        // Note - setAuthor for comment is a depreciated function, so can't provide
        addOption("filter", "Filter id or favorite filter name.");
        addOption("search", "Search JQL query.");
        addOption("link", "Link description. Examples: 'relates to' or 'is related to'. Link id can also be used for deleteLink.");
        addOption("values", "Comma separated list of field values");
        addOption("timeSpent", "Time spent on work using. Example: 3h 30m");

        addIntegerOption("count", "Maximum number of entries to return.", Integer.MAX_VALUE);

        // parameterList.add(new FlaggedOption("votes", JSAP.LONG_PARSER, JSAP.NO_DEFAULT, JSAP.NOT_REQUIRED, JSAP.NO_SHORTFLAG, "votes", "Number of votes."));

        addOption("plugin", "Plugin key or partial key for matching.");
        addOption("version", "Plugin version.");

        addOption("role", "User role in project");

        addOption("group", "Group name");
        addOption("defaultGroup", "Default group to move users on removeGroup action.");
        addOption("userId", "User id for user management and other actions");
        addOption("userFullName", "User name for user management actions");
        addOption("userEmail", "User email for user management actions");
        addOption("userPassword", "User password for user management actions");

        addOption("permissionScheme", "Permission scheme name or id", "Default Permission Scheme");
        addOption("notificationScheme", "Notification scheme name or id");
        addOption("issueSecurityScheme", "Issue security scheme name or id");

        addOption("findReplace", "Find and replace text. Comma separated list of colon separated pairs.");

        addSwitch("autoVersion", "Automatically add versions used in affectsVersions and fixVersions parameters.");
        addSwitch("autoComponent", "Automatically add components used in components parameter.");
        addSwitch("autoGroup", "Groups are automatically added when referenced in add user functions.");
        addSwitch("asVersion", "Interpret values parameter as version values and convert each to the version id.");
        addSwitch("asComponent", "Interpret values parameter as component values and convert each to the component id.");

        addSwitch("copyLinks", "Copy issue links when cloning an issue.");
        addSwitch("copyAttachments", "Copy attachments when cloning an issue.");
    }

    /*
     * Function implementations
     */

    /**
     * Get server information
     */
    public String getServerInfo() throws RemoteException, java.rmi.RemoteException {

        RemoteServerInfo info = service.getServerInfo(token);
        String version = info.getVersion();
        if (version.startsWith("4.2")) {
            version = version.replaceFirst("-b\\d*", ""); // see http://jira.atlassian.com/browse/JRA-23855
        }
        String message = "JIRA version: " + version + ", build: " + info.getBuildNumber() + ", edition: " + info.getEdition() + ", time: "
                + info.getServerTime().getServerTime() + ", time zone: " + info.getServerTime().getTimeZoneId();
        return message;
    }

    /**
     * Create a new issue
     * 
     * @return result message
     * @throws RemoteRestException
     */
    public String createIssue() throws RemoteException, ClientException, RemoteRestException {

        String project = getRequiredString("project");
        String type = getRequiredString("type");
        getRequiredString("summary");
        String parent = getString("parent");
        boolean isSubtask = !parent.equals("");

        project = getRemoteProject(project).getKey();

        RemoteIssue issue = new RemoteIssue();

        issue.setProject(project);
        issue.setType(getRemoteIssueType(type, project, isSubtask).getId());

        issue.setAssignee("-1"); // default for automatic assignment needed prior to 3.13.3

        updateFromParameters(issue, null, false);

        if ("".equals(issue.getReporter())) {
            issue.setReporter(getString("user")); // default to current login user
        }
        if ((issue.getPriority() == null) || "".equals(issue.getPriority())) {
            issue.setPriority("3"); // Major
        }
        String key;
        if (isSubtask) {
            key = createSubtask(issue, parent);
        } else {
            key = service.createIssue(token, issue).getKey();
        }
        return "Issue " + key + " created" + (isSubtask ? " as subtask of " + parent.toUpperCase() : "") + ".";
    }

    /**
     * Create a subtask. Requested project must be same as parent project!
     * 
     * @param issue contains all the values needed to create
     * @return issue key created if successful
     */
    public String createSubtask(final RemoteIssue issue, final String parent) throws ClientException, RemoteRestException, RemoteException {

        String request = "/secure/CreateSubTaskIssueDetails.jspa";

        RemoteIssue parentIssue = getRemoteIssue(parent); // validates parent exists
        RemoteProject project = getRemoteProject(issue.getProject());
        RemoteProject parentProject = getRemoteProject(parentIssue.getProject());
        if (!parentProject.getId().equals(project.getId())) {
            throw new ClientException("Parent issue must be in the same project.");
        }

        if (issue.getReporter() == null) {
            issue.setReporter(getString("user")); // default to logged in user
        }
        if ((issue.getPriority() == null) || "".equals(issue.getPriority())) {
            issue.setPriority("3"); // Major
        }
        Map<String, String> parameters = new HashMap<String, String>(); // this handles encoding etc...
        Map<String, String[]> arrayParameters = new HashMap<String, String[]>(); // this handles encoding etc...

        parameters.put("parentIssueId", parentIssue.getId());
        parameters.put("pid", project.getId());
        parameters.put("issuetype", issue.getType());

        parameters.put("assignee", issue.getAssignee());
        parameters.put("description", issue.getDescription());
        // duedate needs to be specified in the server format provided by the jira.date.time.picker.java.format found only in the jira-application.properties
        // The shipped default is d/MMM/yy.
        // parameters.put("duedate", getCalendarAsString(issue.getDuedate(), "d/MMM/yy"));
        // parameters.put("duedate", getFieldValue(issue, "duedate"));
        parameters.put("duedate", getString("date"));
        parameters.put("environment", issue.getEnvironment());
        parameters.put("priority", issue.getPriority());
        parameters.put("reporter", issue.getReporter());
        parameters.put("summary", issue.getSummary());

        arrayParameters.put("versions", convertToStringIdArray(issue.getAffectsVersions()));
        arrayParameters.put("components", convertToStringIdArray(issue.getComponents()));
        arrayParameters.put("fixVersions", convertToStringIdArray(issue.getFixVersions()));
        // parameters.put("Create", "Create");

        RemoteCustomFieldValue values[] = issue.getCustomFieldValues();
        if (values != null) {
            for (int i = 0; i < values.length; i++) {
                // out.println("key: " + values[i].getCustomfieldId() + ", value: " + values[i].getValues().toString());
                arrayParameters.put(values[i].getCustomfieldId(), values[i].getValues());
            }
        }
        getGuest().restRequest(request, parameters, arrayParameters); // handles the error case with exception

        // TODO: For reporter required, get <div class="errorArea"><h4>Errors</h4><ul><li>Reporter: Reporter is required.</li></ul></div>
        String issueKey = getGuest().getValue("issueKey").trim();
        if (issueKey.equals("") || issueKey.equals(parentIssue.getKey())) { // means that something other than the standard errmsg was in the output, so the
                                                                            // issue was not created
            throw new RemoteRestException("Errors occurred creating issue. Check all required values are specified.");
        }
        return issueKey;
    }

    /**
     * Put parameters for an string array
     */
    public void putParameters(final Map<String, String> parameters, final String key, final String array[]) {
        for (int i = 0; i < array.length; i++) {
            parameters.put(key, array[i]);
        }
    }

    /**
     * Update from parameters - if values is specified, add all changed fields to the values set - a changed field is determined by comparing the existing issue
     * value with the user provided input - if only issue is provided, just update the appropriate field with the user provided input - only add changed fields
     * to values! This helps us avoid a "no change" update. Even though JIRA is good at only flagging changed data in the change log, it will still update the
     * changed date even if no fields have changed - Note: field names can be found in IssueFieldConstants.java file under com.atlassian.jira.issue in the jira
     * source:
     * 
     * <pre>
     * public static final String PROJECT = &quot;project&quot;;
     * public static final String AFFECTED_VERSIONS = &quot;versions&quot;;
     * public static final String ASSIGNEE = &quot;assignee&quot;;
     * public static final String COMPONENTS = &quot;components&quot;;
     * public static final String COMMENT = &quot;comment&quot;;
     * public static final String DESCRIPTION = &quot;description&quot;;
     * public static final String DUE_DATE = &quot;duedate&quot;;
     * public static final String ENVIRONMENT = &quot;environment&quot;;
     * public static final String FIX_FOR_VERSIONS = &quot;fixVersions&quot;;
     * public static final String ISSUE_KEY = &quot;issuekey&quot;;
     * public static final String ISSUE_TYPE = &quot;issuetype&quot;;
     * public static final String THUMBNAIL = &quot;thumbnail&quot;;
     * public static final String ISSUE_LINKS = &quot;issuelinks&quot;;
     * public static final String WORKRATIO = &quot;workratio&quot;;
     * public static final String SUBTASKS = &quot;subtasks&quot;;
     * public static final String ATTACHMENT = &quot;attachment&quot;;
     * public static final String PRIORITY = &quot;priority&quot;;
     * public static final String REPORTER = &quot;reporter&quot;;
     * public static final String SECURITY = &quot;security&quot;;
     * public static final String SUMMARY = &quot;summary&quot;;
     * public static final String TIMETRACKING = &quot;timetracking&quot;;
     * public static final String CREATED = &quot;created&quot;;
     * public static final String UPDATED = &quot;updated&quot;;
     * public static final String STATUS = &quot;status&quot;;
     * public static final String RESOLUTION = &quot;resolution&quot;;
     * </pre>
     * 
     * Comment cannot be added with an update :(.
     * 
     * @param issue (required) existing RemoteIssue object
     * @param values (optional) set to add RemoteFieldValues to
     * @param forceUpdate - if true, update issue even if values match current values. Normally false except for progressIssue to cover up problems with 3.13.x
     *            requiring fields to be present (fixed in 4.x)
     * @throws RemoteException
     * @throws ClientException
     * @throws RemoteRestException
     */
    public void updateFromParameters(RemoteIssue issue, Set<RemoteFieldValue> values, final boolean forceUpdate) throws RemoteException, ClientException,
            RemoteRestException {

        String project = issue.getProject(); // project key may be needed later

        if (jsapResult.userSpecified("summary")) {
            String value = findReplace(convertNewLine(getString("summary")));
            if (values != null) {
                if (forceUpdate || !value.equals(issue.getSummary())) {
                    values.add(new RemoteFieldValue("summary", new String[] {value}));
                }
            } else {
                issue.setSummary(value);
            }
        }
        if (jsapResult.userSpecified("type")) {
            String value = getRemoteIssueType(getString("type"), project).getId();
            if (values != null) {
                if (forceUpdate || !value.equals(issue.getType())) {
                    values.add(new RemoteFieldValue("issuetype", new String[] {value}));
                }
            } else {
                issue.setType(value);
            }
        }
        if (jsapResult.userSpecified("priority")) {
            String value = getRemotePriority(getString("priority")).getId();
            if (values != null) {
                if (forceUpdate || !value.equals(issue.getPriority())) {
                    values.add(new RemoteFieldValue("priority", new String[] {value}));
                }
            } else {
                issue.setPriority(value);
            }
        }
        if (jsapResult.userSpecified("reporter")) {
            String value = getString("reporter");
            if (values != null) {
                if (forceUpdate || !value.equals(issue.getReporter())) {
                    values.add(new RemoteFieldValue("reporter", new String[] {value}));
                }
            } else {
                issue.setReporter(value);
            }
        }
        if (jsapResult.userSpecified("assignee")) {
            String value = getString("assignee");
            if (values != null) {
                if (forceUpdate || !value.equals(issue.getAssignee())) {
                    values.add(new RemoteFieldValue("assignee", new String[] {value}));
                }
            } else {
                issue.setAssignee(value);
            }
        }
        if (jsapResult.userSpecified("description") || jsapResult.userSpecified("file")) {
            String value = findReplace(convertNewLine(getString("description")));
            if (jsapResult.userSpecified("file")) {
                value = value + CliUtils.removeInvalidXmlChars(findReplace(getFileAsString(getString("file"), getString("encoding"))));
            }
            if (values != null) {
                if (forceUpdate || !value.equals(issue.getDescription())) {
                    values.add(new RemoteFieldValue("description", new String[] {value}));
                }
            } else {
                issue.setDescription(value);
            }
        }

        /*
         * duedate needs to be specified in the server format provided by the jira.date.time.picker.java.format found only in the jira-application.properties
         * file. The shipped default is d/MMM/yy. if null update checking is required, dateFormat must be provided to do the comparison
         */
        if (jsapResult.userSpecified("date")) {
            // String value = getDateAsString(getDate(getString("date"), getString("dateFormat")), "d/MMM/yy");
            String value = getString("date"); // use exactly as provided
            if (value.equals("")) {
                // ignore blank value - don't try to validate it is in correct format!
            } else {
                if (values != null) {
                    // For update, in order to check against existing value, need the dateFormat to do the comparison
                    String dateFormat = getString("dateFormat");
                    if (dateFormat.equals("")) {
                        dateFormat = null; // use system default
                    }
                    if (forceUpdate || !value.equals(getCalendarAsString(issue.getDuedate(), dateFormat))) {
                        // out.println("value: " + value + ", issue: " + getCalendarAsString(issue.getDuedate(),
                        // getString("dateFormat")));
                        values.add(new RemoteFieldValue("duedate", new String[] {value}));
                    }
                } else {
                    issue.setDuedate(getCalendar(getString("date"), getString("dateFormat")));
                }
            }
        }
        if (jsapResult.userSpecified("environment")) {
            String value = findReplace(getString("environment"));
            if (values != null) {
                if (forceUpdate || !value.equals(issue.getEnvironment())) {
                    values.add(new RemoteFieldValue("environment", new String[] {value}));
                }
            } else {
                issue.setEnvironment(value);
            }
        }
        // if (jsapResult.userSpecified("security")) {
        // String value = findReplace(getString("security"));
        // String value = "test";
        // if (forceUpdate || values != null) {
        // values.add(new RemoteFieldValue("security", new String[] {value}));
        // }
        // }

        // Votes do not work
        // if (jsapResult.userSpecified("votes")) {
        // Long value = jsapResult.getLong("votes");
        // if (values != null) {
        // if (!value.equals(issue.getVotes())) {
        // values.add(new RemoteFieldValue("votes", new String[] {value.toString()}));
        // }
        // } else {
        // issue.setVotes(value);
        // }
        // }
        if (jsapResult.userSpecified("components")) {
            updateComponents(issue, values, project, forceUpdate);
        }
        if (jsapResult.userSpecified("affectsVersions")) {
            updateVersions("affectsVersions", issue, values, project, forceUpdate);
        }
        if (jsapResult.userSpecified("fixVersions")) {
            updateVersions("fixVersions", issue, values, project, forceUpdate);
        }
        if (jsapResult.userSpecified("custom")) {
            updateCustomFields(issue, values, forceUpdate);
        }
        if (jsapResult.userSpecified("resolution")) {
            updateResolution(values); // resolution is only valid for a workflow progression
        }
    }

    /**
     * Update components value in issue or values. Parameter is split and each entry is validated against the project's components.
     * 
     * @param issue - issue to be updated
     * @param values - value set to be updated
     * @param project - project key
     * @throws RemoteException
     * @throws ClientException - invalid component
     * @throws RemoteRestException
     */
    public void updateComponents(RemoteIssue issue, Set<RemoteFieldValue> values, final String project, final boolean forceUpdate) throws RemoteException,
            ClientException, RemoteRestException {

        String commaSeparated = getString("components");
        String components[] = commaSeparated.equals("") ? new String[0] : commaSeparated.split(",");
        RemoteComponent remoteComponents[] = new RemoteComponent[components.length];
        String ids[] = new String[components.length];
        boolean doValidation = true;
        for (int count = 1; (count <= 2) && doValidation; count++) { // loop twice if new component is added
            doValidation = false;
            for (int i = 0; i < components.length; i++) {
                String component = components[i].trim();
                // out.println("component: " + components[i]);
                RemoteComponent list[] = service.getComponents(token, project);
                int index = findEntityIndex(component, list);
                if (index >= 0) {
                    remoteComponents[i] = list[index];
                    ids[i] = list[index].getId();
                    // out.println("Ids: " + ids[i] + ", long: " + new Long(ids[i]));
                } else {
                    if (jsapResult.userSpecified("autoComponent")) {
                        String message = addComponent(getRemoteProject(project), component, "", "");
                        System.out.println(message);
                        doValidation = true; // need to get the version id that was added
                    } else {
                        throw new ClientException("Component '" + components[i] + "' is not valid for project with key: " + project);
                    }
                }
            }
        }
        if (values != null) {
            RemoteComponent currentComponents[] = issue.getComponents();
            String currentIds[] = new String[currentComponents.length];
            for (int i = 0; i < currentComponents.length; i++) {
                currentIds[i] = currentComponents[i].getId();
            }
            if (forceUpdate || !CliUtils.areArraysSetEquivalent(ids, currentIds)) {
                values.add(new RemoteFieldValue("components", ids));
            }
        } else {
            issue.setComponents(remoteComponents);
        }
    }

    /**
     * Update fix versions or affected versions value in issue or values. Parameter is split and each entry is validated against the projects components.
     * 
     * @param parameter - either fixVersions or affectsVersions
     * @param issue - issue to be updated
     * @param values - value set to be updated
     * @param project - project key
     * @throws RemoteException
     * @throws ClientException - invalid component TODO: There seems to be an Atlassian bug if more than one affects or fix versions entry is provided on issue
     *             create (update is ok)
     */
    public void updateVersions(final String parameter, RemoteIssue issue, Set<RemoteFieldValue> values, final String project, final boolean forceUpdate)
            throws RemoteException, ClientException {

        String commaSeparated = getString(parameter);
        String versions[] = commaSeparated.equals("") ? new String[0] : commaSeparated.split(",");
        RemoteVersion remoteVersions[] = new RemoteVersion[versions.length];
        String ids[] = new String[remoteVersions.length];
        boolean doValidation = true;
        for (int count = 1; (count <= 2) && doValidation; count++) { // loop twice if new version is added
            doValidation = false;
            RemoteVersion list[] = service.getVersions(token, project);
            for (int i = 0; i < versions.length; i++) {
                String version = versions[i].trim();
                int index = findEntityIndex(version, list);
                if (index >= 0) { // found it
                    remoteVersions[i] = list[index];
                    ids[i] = list[index].getId();
                    // out.println("version: " + versions[i] + ", remoteVersion[.]: " + remoteVersions[i] + ", id: " +
                    // ids[i]);
                } else {
                    if (jsapResult.userSpecified("autoVersion")) {
                        String message = addVersion(project, version, null, null);
                        System.out.println(message);
                        doValidation = true; // need to get the version id that was added
                    } else {
                        throw new ClientException("Version '" + version + "' is not valid for project with key: " + project);
                    }
                }
            }
        }
        if (values != null) {
            RemoteVersion currentVersion[];
            if (parameter.equalsIgnoreCase("fixVersions")) {
                currentVersion = issue.getFixVersions();
            } else if (parameter.equalsIgnoreCase("affectsVersions")) {
                currentVersion = issue.getAffectsVersions();
            } else {
                throw new ClientException("Program error - unexpected value of: " + parameter);
            }
            String currentIds[] = new String[currentVersion.length];
            for (int i = 0; i < currentVersion.length; i++) {
                currentIds[i] = currentVersion[i].getId();
            }
            if (forceUpdate || !CliUtils.areArraysSetEquivalent(ids, currentIds)) {
                if (parameter.equals("fixVersions")) {
                    values.add(new RemoteFieldValue("fixVersions", ids));
                } else if (parameter.equals("affectsVersions")) {
                    values.add(new RemoteFieldValue("versions", ids));
                }
            }
        } else {
            if (parameter.equals("fixVersions")) {
                issue.setFixVersions(remoteVersions);
            } else if (parameter.equals("affectsVersions")) {
                issue.setAffectsVersions(remoteVersions);
            }
        }
    }

    /**
     * Update resolution - only applies to workflow transitions
     * 
     * @param values
     * @throws RemoteException
     * @throws ClientException
     */
    public void updateResolution(Set<RemoteFieldValue> values) throws RemoteException, ClientException {
        RemoteResolution list[] = service.getResolutions(token);
        // out.println(list.length);
        // for (RemoteResolution entry: list) {
        // out.println("resolution: " + entry.getName() + ", id: " + entry.getId());
        // }
        String resolution = getString("resolution");
        int index = findEntityIndex(resolution, list);
        String id;
        if (index >= 0) { // found it
            id = list[index].getId();
        } else {
            throw new ClientException("Resolution '" + resolution + "' is not valid");
        }
        if (values != null) {
            values.add(new RemoteFieldValue("resolution", new String[] {id}));
        }
    }

    /**
     * Custom fields - parameter format: --custom "'custom1:value1', custom2:value2, id3:value3" - note that keys are trimmed, values are NOT (blanks are
     * retained). Try to deal with updates that do NOT change the value. Look at current value and compare, if not changed ignore. This handles both the create
     * and update cases. If values is provided, it is the update case
     * 
     * @param issue
     * @param values
     * @throws RemoteException
     * @throws ClientException
     */
    public void updateCustomFields(RemoteIssue issue, Set<RemoteFieldValue> values, final boolean forceUpdate) throws RemoteException, ClientException {

        String custom = getString("custom");
        String entries[] = custom.equals("") ? new String[0] : splitCsvData(custom, ",", "'");

        RemoteCustomFieldValue newCustomFieldValues[] = new RemoteCustomFieldValue[entries.length];

        if (entries.length > 0) {
            warnOnNoAdminPermission(); // check whether custom fields can be validated
        }

        for (int i = 0; i < entries.length; i++) {
            String entry = findReplace(stripQuotes(entries[i], '\''));
            int index = entry.indexOf(':');
            if (index < 0) {
                throw new ClientException("Custom field entry: '" + entry + "' is not in format key:value");
            }
            String id = getRemoteFieldId(entry.substring(0, index).trim());
            String value = entry.substring(index + 1);
            boolean isQuoted = isQuoted(entries[i]); // if value is quoted, keep whatever is there
            if (!isQuoted) {
                value = value.trim();
            }
            value = convertNewLine(value);
            newCustomFieldValues[i] = new RemoteCustomFieldValue();
            newCustomFieldValues[i].setCustomfieldId(id);
            // unquoted, 0 length values are treated as an null custom field value - ie, set to empty string array
            newCustomFieldValues[i].setValues(((value.length() == 0) && !isQuoted) ? new String[0] : new String[] {value});
        }
        // for (int i = 0; i < newCustomFieldValues.length; i++) {
        // out.println("cf[" + i + "]: "+ newCustomFieldValues[i].getCustomfieldId() + ", " +
        // newCustomFieldValues[i].getValues()[0]);
        // }

        // Look for a real change in the field data before adding a change entry
        if (values != null) {
            RemoteCustomFieldValue currentValues[] = issue.getCustomFieldValues();
            // for (int i = 0; i < currentValues.length; i++) {
            // out.println("current cf[" + i + "]: "+ currentValues[i].getKey() + ", " +
            // currentValues[i].getValues()[0]);
            // }

            for (int i = 0; i < newCustomFieldValues.length; i++) {
                if (forceUpdate || !doesCustomFieldValueMatch(newCustomFieldValues[i], currentValues)) {
                    // out.println("not match - cf[" + i + "]: " + newCustomFieldValues[i].getValues()[0]);
                    values.add(new RemoteFieldValue(newCustomFieldValues[i].getCustomfieldId(), newCustomFieldValues[i].getValues()));
                }
            }
        } else {
            issue.setCustomFieldValues(newCustomFieldValues);
        }
    }

    /**
     * Test whether the new custom field value is the same as the current setting
     * 
     * @param newValue - the requested value
     * @param currentValues - all the custom field values for the issue
     * @return true if new value is the same as the current setting
     */
    public boolean doesCustomFieldValueMatch(RemoteCustomFieldValue newValue, final RemoteCustomFieldValue[] currentValues) {
        boolean match = false;
        String id = newValue.getCustomfieldId();
        // need to lookup the specific custom field value in the list
        for (int currentIndex = 0; (currentIndex < currentValues.length); currentIndex++) {
            if (id.equals(currentValues[currentIndex].getCustomfieldId())) {
                // out.println("new: " + newValue.getValues().length + ", " + newValue.getValues()[0]);
                // out.println("old: " + currentValues[currentIndex].getValues().length + ", " + currentValues[currentIndex].getValues()[0]);
                match = CliUtils.areArraysEquivalent(newValue.getValues(), currentValues[currentIndex].getValues());
                // match = newValue.getValues().equals(currentValues[currentIndex].getValues()); // this does not work
                break; // we found corresponding custom field and either matched on value or not
            }
        }
        // out.println("match: " + match);
        return match;
    }

    /**
     * Update from issue - due to bug JRA-8717 (http://jira.atlassian.com/browse/JRA-8717) set all undefined fields to current field setting from issue
     * 
     * @param issue (required) existing RemoteIssue object
     * @param values (required) add all missing values (not already included)
     * @throws RemoteException
     * @throws ClientException
     */
    @Deprecated
    public void updateFromIssue(RemoteIssue issue, Set<RemoteFieldValue> values) throws RemoteException, ClientException {

        if (!isFieldFound("summary", values)) {
            String value = issue.getSummary();
            values.add(new RemoteFieldValue("summary", new String[] {value}));
        }
        if (!isFieldFound("type", values)) {
            String value = issue.getType();
            values.add(new RemoteFieldValue("issuetype", new String[] {value}));
        }
        if (!isFieldFound("priority", values)) {
            String value = issue.getPriority();
            values.add(new RemoteFieldValue("priority", new String[] {value}));
        }
        if (!isFieldFound("reporter", values)) {
            String value = issue.getReporter();
            values.add(new RemoteFieldValue("reporter", new String[] {value}));
        }
        if (!isFieldFound("assignee", values)) {
            String value = issue.getAssignee();
            values.add(new RemoteFieldValue("assignee", new String[] {value}));
        }
        if (!isFieldFound("description", values)) {
            String value = issue.getDescription();
            values.add(new RemoteFieldValue("description", new String[] {value}));
        }
        if (!isFieldFound("duedate", values)) {
            String value = getCalendarAsString(issue.getDuedate(), "d/MMM/yy");
            values.add(new RemoteFieldValue("duedate", new String[] {value}));
        }
        if (!isFieldFound("environment", values)) {
            String value = issue.getEnvironment();
            values.add(new RemoteFieldValue("environment", new String[] {value}));
        }
        if (!isFieldFound("components", values)) {
            RemoteComponent list[] = issue.getComponents();
            values.add(new RemoteFieldValue("components", convertToStringIdArray(list)));
        }
        if (!isFieldFound("affectsVersions", values)) {
            RemoteVersion list[] = issue.getAffectsVersions();
            values.add(new RemoteFieldValue("versions", convertToStringIdArray(list)));
        }
        if (!isFieldFound("fixVersions", values)) {
            RemoteVersion list[] = issue.getAffectsVersions();
            values.add(new RemoteFieldValue("fixVersions", convertToStringIdArray(list)));
        }
        RemoteCustomFieldValue list[] = issue.getCustomFieldValues();
        for (RemoteCustomFieldValue entry : list) {
            if (!isFieldFound(entry.getKey(), values)) {
                values.add(new RemoteFieldValue(entry.getKey(), entry.getValues()));
            }
        }
    }

    /**
     * Convert to remote field value array
     * 
     * @return result message
     */
    public String[] convertToStringIdArray(AbstractNamedRemoteEntity[] list) {
        int length = (list == null) ? 0 : list.length;
        String ids[] = new String[length];
        for (int i = 0; i < length; i++) {
            ids[i] = list[i].getId();
        }
        return ids;
    }

    /**
     * field not included in set
     * 
     * @return result message
     */
    public boolean isFieldFound(String field, Set<RemoteFieldValue> values) {
        // out.println("field: " + field + ", values: " + values);
        for (RemoteFieldValue value : values) {
            if ((value != null) && (field != null) && field.equals(value.getId())) {
                return true;
            }
        }
        return false;
    }

    /**
     * Update an existing issue
     * 
     * @return result message
     * @throws RemoteRestException
     */
    public String updateIssue() throws RemoteException, ClientException, RemoteRestException {

        RemoteIssue issue = getRemoteIssue(getRequiredString("issue")); // check issue exists - otherwise exception is thrown

        Set<RemoteFieldValue> valueSet = new HashSet<RemoteFieldValue>();
        updateFromParameters(issue, valueSet, false);

        return updateIssue(issue.getKey(), valueSet);
    }

    /**
     * Update an existing issue
     * 
     * @return result message
     */
    public String updateIssue(final String issue, final Set<RemoteFieldValue> valueSet) throws RemoteException, ClientException {

        String message;
        if (valueSet.size() > 0) {
            service.updateIssue(token, issue, getRemoteFieldValueArray(valueSet));
            message = "Issue " + issue + " updated.";
        } else {
            message = "There are no field values changed by the parameters specified. Issue " + issue + " not changed.";
        }
        return message;
    }

    /**
     * Link issue
     * 
     * @return result message
     */
    public String linkIssue() throws RemoteException, ClientException, RemoteRestException {

        String link = getRequiredString("link");
        String comment = getString("comment");
        RemoteIssue issue = getRemoteIssue(getRequiredString("issue")); // check issue exists - otherwise exception is thrown
        RemoteIssue toIssue = getRemoteIssue(getRequiredString("toIssue")); // check issue exists - otherwise exception is thrown

        String request = "/secure/LinkExistingIssue.jspa";

        Map<String, String> parameters = new HashMap<String, String>(); // this handles encoding etc...

        parameters.put("id", issue.getId());
        parameters.put("linkKey", toIssue.getKey()); // needs the issue key, not id
        parameters.put("linkDesc", link);
        if (!comment.equals("")) {
            parameters.put("comment", comment);
        }
        getGuest().restRequest(request, parameters); // handles the error case with exception

        return "Issue " + issue.getKey() + " " + link + " " + toIssue.getKey();
    }

    /**
     * Delete an issue link (unlink)
     * 
     * <pre>
     * 1. Get:  /secure/DeleteLink.jspa?id=12645&destId=12646&linkType=10011
     * 2. Post: /secure/DeleteLink.jspa?id=12645&destId=12646&sourceId=&linkType=10011 with post data: confirm=true&Delete=Delete
     * </pre>
     * 
     * @return result message
     */
    public String deleteLink() throws RemoteException, ClientException, RemoteRestException {

        String issue = getRequiredString("issue");
        String toIssue = getRequiredString("toIssue");
        String link = getRequiredString("link");
        String linkId = convertLinkTextToId(link);

        String request = "/secure/DeleteLink.jspa";

        RemoteIssue remoteIssue = getRemoteIssue(issue); // validates issue exists
        RemoteIssue remoteToIssue = getRemoteIssue(toIssue); // validates issue exists

        Map<String, String> parameters = new HashMap<String, String>(); // this handles encoding etc...

        parameters.put("id", remoteIssue.getId());
        parameters.put("destId", remoteToIssue.getId());
        parameters.put("sourceId", "");
        parameters.put("linkType", linkId);
        parameters.put("confirm", "true");
        parameters.put("Delete", "Delete");

        getGuest().restRequest(request, parameters); // handles the error case with exception

        return "'" + link + "' link deleted between " + issue.toUpperCase() + " and " + toIssue.toUpperCase();
    }

    /**
     * Convert link text to an id since delete link requires the id and not the text. If link is already looks like an id, use it
     * 
     * <pre>
     * /secure/admin/jira/ViewLinkTypes.jspa
     * @param link text
     * @return id as a String
     */
    public String convertLinkTextToId(final String link) throws ClientException, RemoteRestException {
        String linkId = CliUtils.matchRegex(link, "\\d{5,}");
        if ("".equals(linkId)) { // not a link id already
            String request = "/secure/ViewLinkTypes.jspa"; // display that shows the text and ids
            Map<String, String> parameters = new HashMap<String, String>(); // this handles encoding etc...
            getGuest().restRequest(request, parameters); // handles the error case with exception

            linkId = getGuest().getValue("linkId", link);
            if ("".equals(linkId)) { // link id not found
                throw new RemoteRestException("Link '" + link + "' could not be found.");
            }
        }
        return link;
    }

    /**
     * Add comment to issue
     * 
     * @return result message
     */
    public String addComment() throws RemoteException, ClientException {

        String issue = getRequiredString("issue").toUpperCase();
        String comment = findReplace(convertNewLine(getString("comment")));
        String role = getString("role"); // role visibility
        String group = getString("group"); // group visibility

        if (jsapResult.userSpecified("file")) {
            comment += findReplace(CliUtils.removeInvalidXmlChars(getFileAsString(getString("file"), getString("encoding"))));
        }
        if ("".equals(comment)) {
            throw new ClientException("A non-blank comment or file parameter is required.");
        }

        RemoteIssue remoteIssue = getRemoteIssue(issue); // check issue exists - otherwise exception is thrown
        RemoteComment remoteComment = new RemoteComment();

        remoteComment.setBody(comment);
        if (!role.equals("")) {
            remoteComment.setRoleLevel(role);
        }
        if (!group.equals("")) {
            remoteComment.setGroupLevel(group);
        }
        // set author is a depreciated function, so don't use
        // if (jsapResult.userSpecified("userId")) {
        // System.out.println("xxxxxxx");
        // remoteComment.setAuthor(getString("userId"));
        // }

        service.addComment(token, remoteIssue.getKey(), remoteComment);
        return "Comment added.";
    }

    /**
     * Add attachment to issue. Mime warning - http://lkamal.blogspot.com/2008/09/eclipse-access-restriction-on-library.html
     * 
     * @return result message
     */
    public String addAttachment() throws RemoteException, ClientException {

        String issue = getRequiredString("issue");
        String fileName = getRequiredString("file");

        String contentType = new MimetypesFileTypeMap().getContentType(fileName);
        boolean textContent = (contentType != null) && contentType.startsWith("text/");

        RemoteIssue remoteIssue = getRemoteIssue(issue); // check issue exists - otherwise exception is thrown

        String name = getString("name");
        if ("".equals(name)) {
            name = (new File(fileName)).getName();
        }
        String[] nameList = {name};

        if (jsapResult.userSpecified("findReplace")) {
            final String[] dataList = {org.apache.axis.encoding.Base64.encode(CliUtils.removeInvalidXmlChars(
                    findReplace(getFileAsString(fileName, getString("encoding")))).getBytes())};
            service.addBase64EncodedAttachmentsToIssue(token, remoteIssue.getKey(), nameList, dataList);
        } else if (!textContent || getString("encoding").equals("binary")) {
            final String[] dataList = {org.apache.axis.encoding.Base64.encode(getFileAsBytes(fileName))};
            service.addBase64EncodedAttachmentsToIssue(token, remoteIssue.getKey(), nameList, dataList);
        } else {
            final String[] dataList = {org.apache.axis.encoding.Base64.encode(getFileAsString(fileName, getString("encoding")).getBytes())};
            service.addBase64EncodedAttachmentsToIssue(token, remoteIssue.getKey(), nameList, dataList);
        }
        return "Attachment '" + name + "' added to issue " + remoteIssue.getKey();
    }

    /**
     * Get comment list
     * 
     * @return result message
     */
    public String getCommentList() throws RemoteException, ClientException {

        String issue = getRequiredString("issue").toUpperCase();

        RemoteIssue remoteIssue = getRemoteIssue(issue); // check issue exists - otherwise exception is thrown
        RemoteComment list[] = service.getComments(token, issue);

        StringBuilder result = new StringBuilder();
        if (list.length > 0) {
            result.append("Id, Author, Created, Update author, Updated, Visibility");
            for (RemoteComment comment : list) {
                result.append("\n" + comment.getId() + ", " + comment.getAuthor() + ", ");
                result.append(doubleQuoteString(getCalendarAsString(comment.getCreated(), getString("dateFormat"))) + ", ");
                if ((comment.getUpdated() != null) && (!comment.getUpdated().equals(comment.getCreated()))
                        || (!comment.getUpdateAuthor().equals(comment.getAuthor()))) {
                    result.append(comment.getUpdateAuthor() + ", ");
                    result.append(doubleQuoteString(getCalendarAsString(comment.getUpdated(), getString("dateFormat"))) + ", ");
                } else {
                    result.append(", , ");
                }
                if (comment.getRoleLevel() != null) {
                    result.append(comment.getRoleLevel());
                } else if (comment.getGroupLevel() != null) {
                    result.append(comment.getGroupLevel());
                }
            }
        }
        String message = "Data for " + list.length + "comments associated with issue: " + remoteIssue.getKey();
        return standardFinish(message, result.toString(), getString("encoding")); // handle file parameter in standard way
    }

    /**
     * Get formatted string of all comments on a page to output or file
     * 
     * @return result message
     */
    public String getComments() throws RemoteException, ClientException {

        String issue = getRequiredString("issue").toUpperCase();

        RemoteIssue remoteIssue = getRemoteIssue(issue); // check issue exists - otherwise exception is thrown
        RemoteComment list[] = service.getComments(token, remoteIssue.getKey());

        StringBuilder result = new StringBuilder();
        if (list.length > 0) {
            String divider = "";
            for (RemoteComment comment : list) {
                result.append(divider);
                result.append("\n" + comment.getAuthor()).append(" added a comment on " + getCalendarAsString(comment.getCreated(), getString("dateFormat")));
                if ((comment.getUpdated() != null) && (!comment.getUpdated().equals(comment.getCreated()))
                        || (!comment.getUpdateAuthor().equals(comment.getAuthor()))) {
                    result.append(" - edited by " + comment.getUpdateAuthor()).append(
                            " on " + getCalendarAsString(comment.getUpdated(), getString("dateFormat")));
                }
                if (comment.getRoleLevel() != null) {
                    result.append(" - visible to " + comment.getRoleLevel());
                } else if (comment.getGroupLevel() != null) {
                    result.append(" - visible to " + comment.getGroupLevel());
                }
                result.append("\n" + comment.getBody());
                divider = "\n- - - - - - - - - - - - - - - - - - - - - - - - - ";
            }
        }

        String message = "Data for " + list.length + "comments";
        return standardFinish(message, result.toString(), getString("encoding")); // handle file parameter in standard way
    }

    /**
     * Get attachment
     * 
     * @return result message
     */
    public String getAttachment() throws RemoteException, ClientException, RemoteRestException {

        String issue = getRequiredString("issue").toUpperCase();
        String file = getRequiredString("file");
        String fileName = getString("name");

        RemoteIssue remoteIssue = getRemoteIssue(issue); // check issue exists - otherwise exception is thrown
        RemoteAttachment list[] = service.getAttachmentsFromIssue(token, remoteIssue.getKey());
        RemoteAttachment foundAttachment = null;

        if (fileName.equals("")) { // if no name provided use the name from the file
            fileName = (new File(file)).getName();
        }
        for (RemoteAttachment attachment : list) {
            // match by name or id
            if (attachment.getFilename().equals(fileName) || attachment.getId().equals(fileName)) {
                if ((foundAttachment == null) || (foundAttachment.getCreated().compareTo(attachment.getCreated()) < 0)) { // get
                    // last
                    // created
                    foundAttachment = attachment;
                }
            }
        }
        if (foundAttachment == null) {
            throw new ClientException("Attachment '" + fileName + "' not found for issue " + remoteIssue.getKey());
        }
        String request = "/secure/attachment/" + foundAttachment.getId() + "/" + foundAttachment.getFilename();

        try {
            URLConnection connection = getGuest().getUrlConnection(request, null);
            byte[] data = getGuest().streamToByteArray(connection.getInputStream(), foundAttachment.getFilesize().intValue());
            writeToFile(data, file, foundAttachment.getMimetype(), getString("encoding"));
        } catch (IOException exception) {
            if (debug) {
                exception.printStackTrace(err);
            }
            throw new ClientException("Error reading attachment data,\n" + exception.toString());
        }
        return "Attachment data written to file: " + file;
    }

    /**
     * List attachments for an issue
     * 
     * @return result message
     */
    public String getAttachmentList() throws RemoteException, ClientException {

        String issue = getRequiredString("issue").toUpperCase();

        RemoteIssue remoteIssue = getRemoteIssue(issue); // check issue exists - otherwise exception is thrown
        RemoteAttachment list[] = service.getAttachmentsFromIssue(token, remoteIssue.getKey());

        StringBuilder result = new StringBuilder();
        result.append("Id, Name, MIME Type, Size, Date Created, Author");

        for (RemoteAttachment attachment : list) {
            result.append("\n" + attachment.getId() + ", ");
            result.append(doubleQuoteString(attachment.getFilename()) + ", ");
            result.append(attachment.getMimetype() + ", ");
            result.append(attachment.getFilesize() + ", ");
            result.append(doubleQuoteString(toString(attachment.getCreated())) + ", " + attachment.getAuthor());
        }
        String message = list.length + " attachments for issue: " + issue;
        return standardFinish(message, result.toString(), getString("encoding")); // handle file parameter in standard way
    }

    /**
     * Get project key from issue key
     */
    public String getProjectKeyFromIssueKey(final String key) throws ClientException {
        int index = key.indexOf('-');
        if (index < 0) {
            throw new ClientException("Issue key '" + key + "' is not valid. Valid format: xxx-nnnn'");
        }
        return key.substring(0, index);
    }

    /**
     * Get project from name, id, or key - note that name is also unique
     */
    public RemoteProject getRemoteProject(final String value) throws ClientException, RemoteException {
        RemoteProject list[] = service.getProjectsNoSchemes(token);
        // look for match by key
        for (int index = 0; index < list.length; index++) {
            if (list[index].getKey().equalsIgnoreCase(value)) {
                return list[index];
            }
        }
        // not found by key value, so try finding by name or id
        int index = findEntityIndex(value, list);
        if (index < 0) {
            throw new ClientException("Project '" + value + "' is not valid");
        }
        return list[index];
    }

    /**
     * Find remote component matching on name or id
     */
    public RemoteComponent getRemoteComponent(final RemoteProject project, final String value) throws ClientException, RemoteException {

        RemoteComponent list[] = service.getComponents(token, project.getKey());

        int index = findEntityIndex(value, list);
        if (index < 0) {
            throw new ClientException("Component '" + value + "' is not valid");
        }
        return list[index];
    }

    /**
     * Get issue by key or id.
     * 
     * @param key
     * @return issue
     */
    RemoteIssue getRemoteIssue(final String key) throws RemoteException {
        RemoteIssue issue;
        boolean isId = key.matches("\\d+"); // if all digits, treat as an id
        try {
            if (isId) {
                issue = service.getIssueById(token, key);
            } else {
                issue = service.getIssue(token, key.toUpperCase());
            }
        } catch (RemoteException exception) {
            out.println("Could not find issue with key" + (isId ? " or id" : "") + ": '" + key + "'");
            throw exception;
        }
        return issue;
    }

    /**
     * Get issue type from name or id value for either regular types or subtask types
     * 
     * @param value
     * @return remote object
     * @throws ClientException
     * @throws RemoteException
     */
    public RemoteIssueType getRemoteIssueType(final String value, final String project) throws ClientException, RemoteException {

        try {
            return getRemoteIssueType(value, project, true); // see if subtask type first
        } catch (ClientException ignore) {
        }
        return getRemoteIssueType(value, project, false); // let this throw exception if not found
    }

    /**
     * Get issue type from name or id value
     * 
     * @param value
     * @return remote object
     * @throws ClientException
     * @throws RemoteException
     */
    public RemoteIssueType getRemoteIssueType(final String value, final String project, final boolean subtask) throws ClientException, RemoteException {
        // RemoteIssueType list[] = service.getIssueTypes(token);
        RemoteProject remoteProject = getRemoteProject(project);
        RemoteIssueType list[];
        if (subtask) {
            list = service.getSubTaskIssueTypesForProject(token, remoteProject.getId());
        } else {
            list = service.getIssueTypesForProject(token, remoteProject.getId());
        }
        int index = findEntityIndex(value, list);
        if (index < 0) {
            throw new ClientException("Issue type '" + value + "' is not valid");
        }
        return list[index];
    }

    /**
     * Test whether an issue is a task or subtask
     * 
     * @return true if issue is a subtask
     */
    public boolean isSubtask(final String issue) throws RemoteException, ClientException {
        return isSubtask(getRemoteIssue(issue));
    }

    /**
     * Test whether an issue is a task or subtask
     * 
     * @return
     */
    public boolean isSubtask(final RemoteIssue remoteIssue) throws RemoteException, ClientException {
        RemoteProject remoteProject = getRemoteProject(remoteIssue.getProject());
        RemoteIssueType list[] = service.getIssueTypesForProject(token, remoteProject.getId());
        int index = findEntityIndex(remoteIssue.getType(), list);
        return (index < 0); // true if not found in the regular issue types for the project
    }

    /**
     * Get issue status from name or id value
     * 
     * @param value
     * @return remote object
     * @throws ClientException
     * @throws RemoteException
     */
    RemoteStatus getRemoteStatus(final String value) throws ClientException, RemoteException {
        RemoteStatus list[] = service.getStatuses(token);
        int index = findEntityIndex(value, list);
        if (index < 0) {
            throw new ClientException("Issue type '" + value + "' is not valid");
        }
        return list[index];
    }

    /**
     * Get priority from name or id value
     * 
     * @param value
     * @return remote object
     * @throws ClientException
     * @throws RemoteException
     */
    RemotePriority getRemotePriority(final String value) throws ClientException, RemoteException {
        RemotePriority list[] = service.getPriorities(token);
        int index = findEntityIndex(value, list);
        if (index < 0) {
            throw new ClientException("Priority '" + value + "' is not valid");
        }
        return list[index];
    }

    /**
     * Get resolution from name or id value
     * 
     * @param value
     * @return remote object
     * @throfws ClientException
     * @throws RemoteException
     */
    RemoteResolution getRemoteResolution(final String value) throws ClientException, RemoteException {
        RemoteResolution list[] = service.getResolutions(token);
        int index = findEntityIndex(value, list);
        if (index < 0) {
            throw new ClientException("Resolution '" + value + "' is not valid");
        }
        return list[index];
    }

    /**
     * Get custom field id - permission sensitive return of custom field id. If not an administrator, just return field as specified. This means user must
     * provide the custom field id and cannot use name
     * 
     * @param field
     * @return string of the fields id or return the field itself if not permitted to see all custom fields for name lookup
     */
    public String getRemoteFieldId(final String field) throws ClientException, RemoteException {
        try {
            return getRemoteField(field).getId();
        } catch (RemotePermissionException exception) {
            if (verbose) {
                out.println("Permission not sufficient to get custom field list.");
            }
        }
        return field;
    }

    /**
     * Get custom field name - permission sensitive return of custom field name. If not an administrator, just return field as specified
     * 
     * @param field
     * @return string of the fields id or return the field itself if not permitted to see all custom fields for name lookup
     */
    public String getRemoteFieldName(final String field) throws ClientException, RemoteException {
        try {
            return getRemoteField(field).getName();
        } catch (RemotePermissionException exception) {
            if (verbose) {
                out.println("Permission not sufficient to get custom field list.");
            }
        }
        return field;
    }

    /**
     * Get custom field by name or id (with customfield_) or raw id (number)
     * 
     * @param field
     * @return field object
     */
    public RemoteField getRemoteField(final String field) throws ClientException, RemoteException {
        RemoteField list[] = service.getCustomFields(token);
        int index = findEntityIndex(field, list);
        if (index < 0) {
            String fieldAsId = "customfield_" + field;
            index = findEntityIndex(fieldAsId, list);
            if (index < 0) {
                throw new RemoteException("Custom field '" + field + "' is not valid");
            }
        }
        return list[index];
    }

    public void warnOnNoAdminPermission() throws com.dolby.jira.net.soap.jira.RemoteException, RemoteException {
        try {
            service.getCustomFields(token);
        } catch (RemotePermissionException exception) {
            out.println("Warning: user is not an administrator. Custom fields cannot be validated. Use correct custom field ids!");
        }
    }

    /**
     * Create a new project
     * 
     * @return result message
     */
    public String createProject() throws RemoteException, ClientException {

        String project = getRequiredString("project");
        String lead = getRequiredString("lead");
        String name = getString("name");
        if ("".equals(name)) {
            name = project; // default the name to be the project key
        }
        project = project.toUpperCase(); // key must be upper case
        RemoteProject remoteProject = new RemoteProject();
        remoteProject.setKey(project);
        remoteProject.setName(name);
        remoteProject.setLead(lead);
        remoteProject.setDescription(getString("description"));
        remoteProject.setIssueSecurityScheme(getScheme(getString("issueSecurityScheme"), "Issue Security", jsapResult.userSpecified("issueSecurityScheme")));
        remoteProject.setNotificationScheme(getScheme(getString("notificationScheme"), "Notification", jsapResult.userSpecified("notificationScheme")));
        remoteProject.setPermissionScheme((RemotePermissionScheme) getScheme(getString("permissionScheme"), "Permission",
                jsapResult.userSpecified("permissionScheme")));
        remoteProject.setUrl(getString("url"));
        service.createProjectFromObject(token, remoteProject);

        return "Project '" + name + "' created with key " + project;
    }

    /**
     * Delete a project
     * 
     * @return result message
     */
    public String deleteProject() throws RemoteException, ClientException {

        String project = getRequiredString("project");

        // Makes the error messaging better if it cannot be found
        try {
            getRemoteProject(project);
        } catch (RemoteException exception) {
            throw exception;
        } catch (ClientException exception) {
            return "Project " + project + " not found.  Request is ignored.";
        }
        service.deleteProject(token, project.toUpperCase());
        return "Project " + project + " deleted.";
    }

    /**
     * Add a version to a project
     * 
     * @return result message
     */
    public String addVersion() throws RemoteException, ClientException {

        String project = getRemoteProject(getRequiredString("project")).getKey();
        String name = getRequiredString("name");
        String after = getString("after");
        Calendar calendar = null;
        if (jsapResult.userSpecified("date")) {
            calendar = getCalendar(getString("date"), getString("dateFormat"));
        }
        return addVersion(project, name, after, calendar);
    }

    /**
     * Add a version to a project
     * 
     * @return result message
     */
    public String addVersion(final String project, final String name, final String after, final Calendar calendar) throws RemoteException, ClientException {

        String afterMessage = "";
        RemoteVersion remoteVersion = new RemoteVersion();
        remoteVersion.setName(name);
        if ((after != null) && !after.equals("")) {
            RemoteVersion afterVersion = getRemoteVersion(project, after, true);
            if (afterVersion != null) {
                remoteVersion.setSequence(afterVersion.getSequence());
                afterMessage = " after version " + after;
            }
        }
        if (calendar != null) {
            remoteVersion.setReleaseDate(calendar);
        }
        service.addVersion(token, project, remoteVersion);

        return "Version '" + name + "' added to project " + project + afterMessage;
    }

    /**
     * Get available workflow steps
     * 
     * @return message
     * @throws RemoteException
     * @throws ClientException
     */
    public String getAvailableSteps() throws RemoteException, ClientException {

        String issue = getRequiredString("issue").toUpperCase();
        RemoteNamedObject[] availableSteps = service.getAvailableActions(token, issue);

        StringBuilder result = new StringBuilder();
        result.append("Id, Name");
        for (RemoteNamedObject step : availableSteps) {
            result.append("\n" + step.getId() + ", " + doubleQuoteString(step.getName()));
        }
        String message = availableSteps.length + " available workflow steps for issue: " + issue;
        return standardFinish(message, result.toString(), getString("encoding")); // handle file parameter in standard way
    }

    /**
     * Progress issue
     * 
     * @return message
     * @throws RemoteException
     * @throws ClientException
     * @throws RemoteRestException
     */
    public String progressIssueInWorkflow() throws RemoteException, ClientException, RemoteRestException {

        String issue = getRequiredString("issue").toUpperCase();
        String step = getRequiredString("step");
        RemoteIssue remoteIssue = getRemoteIssue(issue); // check issue exists - otherwise exception is thrown
        String stepId = getStepId(remoteIssue.getKey(), step);

        Set<RemoteFieldValue> valueSet = new HashSet<RemoteFieldValue>();
        updateFromParameters(remoteIssue, valueSet, true);
        /*
         * for (RemoteFieldValue value : valueSet) { out.println("Id: " + value.getId()); } updateFromIssue(remoteIssue, valueSet); for (RemoteFieldValue value
         * : valueSet) { out.println("Id: " + value.getId()); }
         */
        service.progressWorkflowAction(token, remoteIssue.getKey(), stepId, getRemoteFieldValueArray(valueSet));
        String message = "Successfully progressed issue: " + remoteIssue.getKey() + " with step '" + step + "'";

        try {
            String comment = findReplace(convertNewLine(getString("comment")));
            if (!"".equals(comment)) {
                RemoteComment remoteComment = new RemoteComment();
                remoteComment.setBody(comment);
                service.addComment(token, issue, remoteComment);
            }
        } catch (RemoteException exception) {
            out.println(message); // ensure the progress issue message goes out even if there were problems adding comment
            throw exception;
        }
        return message;
    }

    String getStepId(final String issue, final String step) throws ClientException, RemoteException {
        RemoteNamedObject[] availableActions = service.getAvailableActions(token, issue);
        if (availableActions != null) {
            for (RemoteNamedObject action : availableActions) {
                if (action.getName().equalsIgnoreCase(step) || action.getId().equals(step)) {
                    return action.getId();
                }
            }
        }
        throw new ClientException("Workflow step '" + step + "' is not valid for issue: " + issue);
    }

    /**
     * Get issue
     * 
     * @return result message
     */
    public String getIssue() throws RemoteException, ClientException {

        RemoteIssue issue = getRemoteIssue(getRequiredString("issue"));
        StringBuilder result = new StringBuilder(1000);
        result.append("\n" + CliUtils.prettyOutput("Issue key", getFieldValue(issue, "key")));
        result.append("\n" + CliUtils.prettyOutput("Type", getFieldValue(issue, "type")));
        result.append("\n" + CliUtils.prettyOutput("Status", getFieldValue(issue, "status")));
        result.append("\n" + CliUtils.prettyOutput("Priority", getFieldValue(issue, "priority")));
        result.append("\n" + CliUtils.prettyOutput("Assignee", getFieldValue(issue, "assignee")));
        result.append("\n" + CliUtils.prettyOutput("Reporter", getFieldValue(issue, "reporter")));
        result.append("\n" + CliUtils.prettyOutput("Created", getFieldValue(issue, "created")));
        result.append("\n" + CliUtils.prettyOutput("Updated", getFieldValue(issue, "updated")));
        result.append("\n" + CliUtils.prettyOutput("Due date", getFieldValue(issue, "duedate")));
        result.append("\n" + CliUtils.prettyOutput("Summary", getFieldValue(issue, "summary")));
        result.append("\n" + CliUtils.prettyOutput("Description", getFieldValue(issue, "description")));
        result.append("\n" + CliUtils.prettyOutput("Components", getFieldValue(issue, "components")));
        result.append("\n" + CliUtils.prettyOutput("Affects versions", getFieldValue(issue, "affectsversions")));
        result.append("\n" + CliUtils.prettyOutput("Fix versions", getFieldValue(issue, "fixversions")));
        result.append("\n" + CliUtils.prettyOutput("Environment", getFieldValue(issue, "environment")));
        result.append("\n" + CliUtils.prettyOutput("Resolution", getFieldValue(issue, "resolution")));
        result.append("\n" + CliUtils.prettyOutput("Votes", getFieldValue(issue, "votes")));
        RemoteCustomFieldValue values[] = issue.getCustomFieldValues();
        if (values != null) {
            for (RemoteCustomFieldValue value : values) {
                if (value != null) {
                    String valueString = null;
                    if (jsapResult.userSpecified("asVersion")) { // no easy way to know this is a version custom field
                        valueString = getFieldValueAsVersionString(issue, value.getValues());
                    }
                    if (valueString == null) {
                        valueString = toString(value.getValues());
                    }
                    // result.append("\n" + prettyOutput(getRemoteFieldName(value.getCustomfieldId()), toString(value.getValues())));
                    result.append("\n" + CliUtils.prettyOutput(getRemoteFieldName(value.getCustomfieldId()), valueString));
                }
            }
        }

        String message = "Data for issue: " + getFieldValue(issue, "key");
        return standardFinish(message, result.toString(), getString("encoding")); // handle file parameter in standard way
    }

    /**
     * Set field value - must be a custom field. This provides a way to set long text custom fields. Only fields on edit screen can be changed :(. see
     * http://www.nabble.com/Jira-SOAP-interface-and-remotely-updating-custom-fields-td22421561.html and http://jira.atlassian.com/browse/JRA-19085
     * 
     * @return field value or message
     */
    public String setFieldValue() throws RemoteException, ClientException {

        String field = getRequiredString("field");
        String value = getString("values"); // TODO - should we add values ???? comma separated?

        String valuesArray[];
        if (!jsapResult.userSpecified("values")) {
            if (!jsapResult.userSpecified("file")) {
                throw new ClientException("Either value or file parameters must be specified");
            }
            valuesArray = new String[] {CliUtils.removeInvalidXmlChars(getFileAsString(getString("file"), getString("encoding")))};
        } else {
            valuesArray = value.equals("") ? new String[0] : splitCsvData(value, ",", "'");
        }
        RemoteIssue issue = getRemoteIssue(getRequiredString("issue"));
        // TODO - complete this
        // RemoteIssue remoteIssue = getRemoteIssue(issue); // Verifies issue is valid

        if (jsapResult.userSpecified("asVersion")) {
            for (int i = 0; i < valuesArray.length; i++) {
                valuesArray[i] = getRemoteVersion(issue.getProject(), valuesArray[i], true).getId();
            }
        } else if (jsapResult.userSpecified("asComponent")) {
            for (int i = 0; i < valuesArray.length; i++) {
                valuesArray[i] = getRemoteComponent(getRemoteProject(issue.getProject()), valuesArray[i]).getId();
            }
        }

        Set<RemoteFieldValue> valueSet = new HashSet<RemoteFieldValue>();
        warnOnNoAdminPermission(); // check whether custom fields can be validated
        String remoteFieldId = getRemoteFieldId(field); // validates this is a valid field if permitted
        RemoteCustomFieldValue newValue = new RemoteCustomFieldValue();
        newValue.setCustomfieldId(remoteFieldId);
        newValue.setValues(valuesArray);
        if (!doesCustomFieldValueMatch(newValue, issue.getCustomFieldValues())) {
            valueSet.add(new RemoteFieldValue(remoteFieldId, valuesArray));
        }
        return updateIssue(issue.getKey(), valueSet);
    }

    /**
     * Get field value
     * 
     * @return field value or message
     */
    public String getFieldValue() throws RemoteException, ClientException {

        String field = getRequiredString("field");
        RemoteIssue issue = getRemoteIssue(getRequiredString("issue"));

        String result = getFieldValue(issue, field);

        // TODO - handle file case
        String message = "Issue " + issue.getKey() + " has field '" + field + "' with value: ";
        return standardFinish(message, result, getString("encoding")); // handle file parameter in standard way
    }

    /**
     * Get field value formatted for output
     * 
     * @return field value or error message
     */
    public String getFieldValue(final RemoteIssue issue, final String inField) throws RemoteException, ClientException {

        String field = inField.toLowerCase();
        if (field.equals("type") || field.equals("issuetype")) {
            String typeString = "sub-task(" + issue.getType() + ")";
            // Subtasks are not valid issue types, so can't get their name - see
            // http://jira.atlassian.com/browse/JRA-6896
            try {
                typeString = getRemoteIssueType(issue.getType(), issue.getProject()).getName() + " (" + issue.getType() + ")";
            } catch (Exception ignore) {
            }
            return typeString;
        } else if (field.equals("key")) {
            return issue.getKey() + " (" + issue.getId() + ")";
        } else if (field.equals("status")) {
            return getRemoteStatus(issue.getStatus()).getName() + " (" + issue.getStatus() + ")";
        } else if (field.equals("priority")) {
            return ((issue.getPriority() == null) ? "" : toString(getRemotePriority(issue.getPriority()).getName() + " (" + issue.getPriority() + ")"));
        } else if (field.equals("assignee")) {
            return toString(issue.getAssignee());
        } else if (field.equals("reporter")) {
            return toString(issue.getReporter());
        } else if (field.equals("created")) {
            return toString(issue.getCreated());
        } else if (field.equals("updated")) {
            return toString(issue.getUpdated());
        } else if (field.equals("duedate")) {
            return toString(issue.getDuedate());
        } else if (field.equals("summary")) {
            return toString(issue.getSummary());
        } else if (field.equals("description")) {
            return toString(issue.getDescription());
        } else if (field.equals("components")) {
            return toString(issue.getComponents());
        } else if (field.equals("affectsversions")) {
            return toString(issue.getAffectsVersions());
        } else if (field.equals("fixversions")) {
            return toString(issue.getFixVersions());
        } else if (field.equals("environment")) {
            return toString(issue.getEnvironment());
        } else if (field.equals("resolution")) {
            return ((issue.getResolution() == null) ? "" : toString(getRemoteResolution(issue.getResolution()).getName() + " (" + issue.getResolution() + ")"));
        } else if (field.equals("status")) {
            return toString(issue.getStatus());
        } else if (field.equals("votes")) {
            return toString(issue.getVotes());
        } else {
            String remoteFieldId = getRemoteFieldId(field); // this tries to validates it is a custom field if an administrator
            RemoteCustomFieldValue values[] = issue.getCustomFieldValues();
            if (values != null) {
                for (RemoteCustomFieldValue value : values) {
                    if (remoteFieldId.equals(value.getCustomfieldId())) { // value matches an id
                        if (jsapResult.userSpecified("asVersion")) {
                            String result = getFieldValueAsVersionString(issue, value.getValues());
                            if (result != null) { // all values are versions and there is at least one
                                return result; // return the version string
                            }
                            // otherwise fall through to normal processing
                        }
                        return toString(value.getValues());
                    }
                }
            }
            return ""; // value not defined for this valid custom field
        }
    }

    /**
     * If the field versionList contains values that are all valid version ids or names, then return a string representing the versions
     * 
     * @param issue
     * @param versionList
     * @return version string or null if some value was not a valid version
     * @throws RemoteException
     * @throws ClientException
     */
    public String getFieldValueAsVersionString(final RemoteIssue issue, final String[] versionList) throws RemoteException, ClientException {
        StringBuilder builder = new StringBuilder();
        boolean isVersionValue = true;
        for (String versionName : versionList) {
            RemoteVersion remoteVersion = getRemoteVersion(issue.getProject(), versionName, false);
            isVersionValue = isVersionValue && (remoteVersion != null);
            if (isVersionValue) {
                builder.append(", '").append(remoteVersion.getName()).append("' (").append(remoteVersion.getId()).append(")");
            }
            // out.println("Version name: " + versionName + ", isVersionValue: " + isVersionValue);
        }
        if (isVersionValue && (builder.length() >= 2)) { // all values are versions and there is at least one
            return builder.toString().substring(2); // return the version string
        }
        return null;
    }

    /**
     * Convert various values to appropriate string for display purposes including arrays to comma separated lists
     * 
     * @param value
     * @return
     */
    public String toString(final Object value) {
        String result = "";
        if (value != null) {
            // out.println("toString: " + value.getClass().getName() + ", value: " + value);
            if (value instanceof String) {
                result = (String) value;
            } else if (value instanceof String[]) {
                String list[] = (String[]) value;
                for (String string : list) {
                    if (string != null) {
                        result += ", '" + string + "'";
                    }
                }
                if (!result.equals("")) {
                    result = result.substring(2); // remove leading comma
                }
            } else if (value instanceof Calendar) {
                result = getCalendarAsString((Calendar) value, getString("dateFormat"));
            } else if (value instanceof Date) {
                result = getDateAsString((Date) value, getString("dateFormat"));
            } else if (value instanceof Long) {
                result = ((Long) value).toString();
            } else if (value instanceof AbstractNamedRemoteEntity) {
                result = ((AbstractNamedRemoteEntity) value).getName() + "(" + ((AbstractNamedRemoteEntity) value).getId() + ")";
            } else if (value instanceof AbstractNamedRemoteEntity[]) {
                AbstractNamedRemoteEntity list[] = (AbstractNamedRemoteEntity[]) value;
                for (AbstractNamedRemoteEntity entry : list) {
                    String string = toString(entry);
                    if (string != null && !string.equals("")) {
                        result += ", '" + (entry).getName() + "'(" + (entry).getId().toString() + ")";
                    }
                }
                if (!result.equals("")) {
                    result = result.substring(2); // remove leading comma
                }
            }
        }
        return result;
    }

    /**
     * Get project list
     * 
     * @return result message
     */
    public String getProjectList() throws RemoteException, ClientException {

        RemoteProject list[] = service.getProjectsNoSchemes(token);

        StringBuilder result = new StringBuilder();
        result.append("Key, Id, Name");
        for (RemoteProject entry : list) {
            result.append("\n" + entry.getKey() + ", " + entry.getId() + ", " + doubleQuoteString(entry.getName()));
        }
        String message = list.length + " projects";
        return standardFinish(message, result.toString(), getString("encoding")); // handle file parameter in standard way
    }

    /**
     * Get issue list for a filter
     * 
     * @return result message
     */
    public String getIssueList() throws RemoteException, ClientException {

        RemoteFilter filter = jsapResult.userSpecified("filter") ? getRemoteFilter(getString("filter")) : null;
        String search = null;
        if (filter == null) {
            search = getString("search");
            if (search.equals("")) {
                throw new ClientException("Either filter or search parameters must be specified.");
            } else {
                String version = service.getServerInfo(token).getVersion();
                if (version.substring(0, 3).compareTo("4.1") < 0) {
                    throw new ClientException("JIRA 4.1 or higher is required to use the search parameter.");
                }
            }
        }

        String dateFormat = getString("dateFormat");

        int count = 0;
        int limitCount = 250;
        boolean more = true;
        StringBuilder result = new StringBuilder();
        result.append("Key, Id, Project, Status, Priority, Assignee, Reporter, Created, Date due");

        for (int offset = 0; more; offset += limitCount) {

            RemoteIssue list[] = (filter != null) ? service.getIssuesFromFilterWithLimit(token, filter.getId(), offset, limitCount) : service
                    .getIssuesFromJqlSearch(token, search, limitCount);
            // RemoteIssue list[] = service.getIssuesFromFilterWithLimit(token, filter.getId(), offset, limitCount);
            more = list.length == limitCount;

            for (RemoteIssue entry : list) {
                count++;
                result.append("\n" + entry.getKey() + ", " + entry.getId() + ", " + entry.getProject() + ", ");
                result.append(entry.getStatus() + ", " + entry.getPriority() + ", " + entry.getAssignee() + ", ");
                result.append(entry.getReporter() + ", " + doubleQuoteString(getCalendarAsString(entry.getCreated(), dateFormat)) + ", ");
                result.append(doubleQuoteString(getCalendarAsString(entry.getDuedate(), dateFormat)));
            }
        }
        String message = count + " issues";
        return standardFinish(message, result.toString(), getString("encoding")); // handle file parameter in standard way
    }

    /**
     * Get filter by id or name
     * 
     * @throws RemoteException
     * @throws com.dolby.jira.net.soap.jira.RemoteException
     * @throws RemoteAuthenticationException
     * @throws RemotePermissionException
     * @throws ClientException
     */
    public RemoteFilter getRemoteFilter(final String value) throws RemotePermissionException, RemoteAuthenticationException,
            com.dolby.jira.net.soap.jira.RemoteException, RemoteException, ClientException {
        RemoteFilter list[] = service.getFavouriteFilters(token);

        int index = findEntityIndex(value, list);
        if (index < 0) {
            throw new ClientException("Filter '" + value + "' is not valid");
        }
        return list[index];
    }

    /**
     * Get remote user
     * 
     * @return user or throw error if not found
     */
    // public RemoteUser getRemoteUser(final String userId) throws java.rmi.RemoteException {
    // RemoteUser result = service.getUser(token, userId);
    // if (result == null) {
    // throw new RemoteException(userId + " is not a valid user.");
    // }
    // return result;
    // }

    /**
     * Get remote group
     * 
     * @return group or throw error if not found
     */
    // public RemoteGroup getRemoteGroup(final String group) throws java.rmi.RemoteException {
    // RemoteGroup result = null;
    // try {
    // result = service.getGroup(token, group);
    // } catch (RemoteValidationException exception) {
    // ignore
    // }
    // if (result == null) {
    // throw new RemoteException(group + " is not a valid group.");
    // }
    // return result;
    // }

    /**
     * List users in a group
     * 
     * @return result message
     */
    public String getUserList() throws RemoteException, ClientException {

        String group = getRequiredString("group");

        RemoteGroup remoteGroup = service.getGroup(token, group);

        StringBuilder result = new StringBuilder();
        result.append("User, Name, Email");
        RemoteUser[] users = remoteGroup.getUsers();

        for (RemoteUser user : users) {
            result.append("\n" + user.getName() + ", " + doubleQuoteString(user.getFullname()) + ", " + user.getEmail());
        }
        String message = remoteGroup.getUsers().length + " users in " + group;
        return standardFinish(message, result.toString(), getString("encoding")); // handle file parameter in standard way
    }

    /**
     * Add component
     * 
     * @param remote project
     * @param component name
     * @param component description
     * @param component lead
     * @return result message
     */
    public String addComponent(final RemoteProject project, final String name, final String description, final String lead) throws RemoteException,
            RemoteRestException, ClientException {

        String message;
        if (isComponentDefined(name, project)) {
            message = "'" + name + "' is already a component for project " + project.getKey() + ". Add request ignored.";
        } else {
            String request = "/secure/project/AddComponent.jspa";

            Map<String, String> parameters = new HashMap<String, String>(); // this handles encoding etc...
            parameters.put("pid", project.getId());
            parameters.put("name", name); // component name
            parameters.put("description", description);
            parameters.put("lead", lead);
            getGuest().restRequest(request, parameters); // handles the error case with exception

            // verify the component really got added
            if (!isComponentDefined(name, project)) {
                if (debug) {
                    out.println("Response from server: " + getGuest().jiraResponse);
                }
                throw new RemoteRestException("Component '" + name + "' not created in project: " + project.getKey()
                        + ".  Reason is unknown. Use --debug for more details.");
            }
            message = "Component '" + name + "' added to project: " + project.getKey();
        }
        return message;
    }

    /**
     * Delete component
     * 
     * @return result message
     */
    public String deleteComponent() throws RemoteException, RemoteRestException, ClientException {

        RemoteProject project = getRemoteProject(getRequiredString("project")); // validate
        String name = getRequiredString("name");

        return deleteComponent(project, name);
    }

    /**
     * Delete component
     * 
     * @return result message
     */
    public String deleteComponent(final RemoteProject project, final String name) throws RemoteException, RemoteRestException, ClientException {

        String message;
        if (!isComponentDefined(name, project)) {
            message = "'" + name + "' is not a component for project " + project.getKey() + ". Delete request ignored.";
        } else {
            RemoteComponent component = getRemoteComponent(project, name); // validates component or issues exception

            String request = "/secure/project/DeleteComponent.jspa";

            Map<String, String> parameters = new HashMap<String, String>(); // this handles encoding etc...
            // parameters.put("pid", project.getId());
            parameters.put("cid", component.getId());
            parameters.put("confirm", "false");
            parameters.put("action", "remove");

            getGuest().restRequest(request, parameters); // handles the error case with exception
            // verify the component really is not there anymore
            if (isComponentDefined(name, project)) {
                if (debug) {
                    out.println("Response from server: " + getGuest().jiraResponse);
                }
                throw new RemoteRestException("Component '" + name + "' not deleted from project: " + project.getKey()
                        + ". Reason is unknown. Use --debug more details.");
            }
            message = "Component '" + name + "' deleted from project: " + project.getKey();
        }
        return message;
    }

    /**
     * is name in list
     * 
     * @return true if name is a component in the project
     */
    public boolean isComponentDefined(final String name, final RemoteProject project) throws RemotePermissionException, RemoteAuthenticationException,
            com.dolby.jira.net.soap.jira.RemoteException, RemoteException {
        boolean found = false;
        RemoteComponent list[] = service.getComponents(token, project.getKey());
        for (int i = 0; (i < list.length) && !found; i++) {
            found = list[i].getName().equalsIgnoreCase(name); // case insensitive search
        }
        return found;
    }

    /**
     * Copy components - unfortunately, description and lead are not available to copy !!!
     * 
     * @return result message
     */
    public String copyComponents() throws RemoteException, RemoteRestException, ClientException {

        RemoteProject project = getRemoteProject(getRequiredString("project")); // validate
        RemoteProject toProject = getRemoteProject(getRequiredString("toProject")); // validate

        RemoteComponent list[] = service.getComponents(token, project.getKey());
        RemoteComponent toList[] = service.getComponents(token, toProject.getKey());

        String commaSeparated = getString("components");
        boolean doAll = commaSeparated.equals("");
        String components[] = doAll ? new String[list.length] : commaSeparated.split(",");

        boolean success = false; // no successful copies yet
        StringBuilder result = new StringBuilder();
        for (int i = 0; i < components.length; i++) {
            // out.println("component: " + components[i]);
            String component = doAll ? list[i].getName() : components[i];
            int index = doAll ? i : findEntityIndex(component, list);
            if (index >= 0) {
                if (findEntityIndex(component, toList) < 0) { // not found in target project
                    result.append(addComponent(toProject, list[index].getName(), "", ""));
                    result.append("\n");
                    success = true; // at least one successful copy
                } else {
                    result.append("Component '" + components[i] + "' already defined for project: " + toProject.getKey());
                    result.append("\n");
                }
            } else {
                result.append("Component '" + components[i] + "' not found in project: " + project.getKey());
                result.append("\n");
            }
        }
        if (!success) {
            out.println(result.toString());
            throw new RemoteRestException("No components copied.");
        }
        return result.toString();
    }

    public String addWork() throws ClientException, RemoteException {
        String issue = getRemoteIssue(getRequiredString("issue")).getKey();
        String user = getString("userId");
        if (user.equals("")) {
            user = getString("user"); // if userId is not set, use logged in user
        }
        Calendar startDate = jsapResult.userSpecified("date") ? getCalendar(getRequiredString("date"), getString("dateFormat")) : Calendar.getInstance();

        RemoteWorklog work = new RemoteWorklog();
        work.setAuthor(user); // THIS HAS NO AFFECT, it is the logged in user always
        work.setUpdateAuthor(user);
        work.setComment(getString("comment"));
        work.setStartDate(startDate);
        // work.setCreated(created);
        work.setTimeSpent(getRequiredString("timeSpent"));
        service.addWorklogAndRetainRemainingEstimate(token, issue, work);
        return "Work log added for issue " + issue + " for user " + user + ".";
    }

    /**
     * Clone issue CloneIssueDetails.jspa?summary=xxxx&id=11111&Create=Create
     */
    public String cloneIssue() throws RemoteException, ClientException, RemoteRestException {

        String issue = getRemoteIssue(getRequiredString("issue")).getKey();
        boolean copyLinks = jsapResult.userSpecified("copyLinks");
        boolean copyAttachments = jsapResult.userSpecified("copyAttachments");

        String request = "/secure/CloneIssueDetails.jspa";
        RemoteIssue remoteIssue = getRemoteIssue(issue); // validates issue exists

        Map<String, String> parameters = new HashMap<String, String>(); // this handles encoding etc...

        parameters.put("id", remoteIssue.getId());
        parameters.put("summary", remoteIssue.getSummary());
        if (copyLinks) {
            parameters.put("cloneLinks", "true");
        }
        if (copyAttachments) {
            parameters.put("cloneAttachments", "true");
        }
        parameters.put("Create", "Create");

        getGuest().restRequest(request, parameters); // handles the error case with exception
        String clonedIssue = getGuest().getValue("issueKey");
        if (clonedIssue.equals("")) {
            throw new RemoteRestException("Clone of issue " + issue + " failed.");
        }
        return "Issue " + issue.toUpperCase() + " cloned to " + getGuest().getValue("issueKey");
    }

    /*
     * Help functions
     */

    /**
     * Find entity that matches element from list either by name or id - match ignoring case and trimming leading and trailing blanks
     * 
     * @param value - value to find, either a name or id of an entity
     * @param list - entities to look through
     * @return index >= 0 if found, -1 otherwise
     */
    public int findEntityIndex(final String entity, final AbstractNamedRemoteEntity list[]) {
        if ((entity != null) && (list != null)) {
            String value = entity.trim();
            for (int index = 0; index < list.length; index++) {
                if ((list[index] != null) && (value.equalsIgnoreCase(list[index].getName()) || value.equals(list[index].getId()))) {
                    return index;
                }
            }
        }
        return -1;
    }

    /**
     * Find entity that matches element from list either by name or id - match ignoring case and trimming blanks
     * 
     * @param value - value to find, either a name or id of an entity
     * @param list - entities to look through
     * @return index >= 0 if found, -1 otherwise
     */
    public int findEntityIndex(final String entity, final RemoteScheme list[]) {
        String value = entity.trim();
        if ((value != null) && (list != null)) {
            for (int index = 0; index < list.length; index++) {
                if ((list[index] != null) && (value.equalsIgnoreCase(list[index].getName()) || value.equals(list[index].getId()))) {
                    return index;
                }
            }
        }
        return -1;
    }

    /**
     * Get user
     * 
     * @param userName
     * @return remoteUser
     */
    public RemoteUser getUser(final String userName) throws RemoteException, ClientException {
        if (remoteUser == null) {
            remoteUser = service.getUser(token, userName);
        }
        return remoteUser;
    }

    /**
     * Get version by name or id
     * 
     * @param version name or id
     * @param project key
     * @param required - if true, throw exception if not found
     * @return remoteVersion
     */
    public RemoteVersion getRemoteVersion(final String project, final String name, boolean required) throws RemoteException, ClientException {

        RemoteVersion result = null;
        RemoteVersion[] list = service.getVersions(token, project);
        int index = findEntityIndex(name, list);
        if (index >= 0) {
            result = list[index];
        }
        if (required && (result == null)) {
            throw new RemoteException("Version '" + name + "' not found");
        }
        return result;
    }

    /**
     * Get scheme by name
     * 
     * @param name - scheme name
     * @param type - type of scheme (issueSecurityScheme, notificationScheme, permissionScheme)
     * @param required - if true, throw exception if not found
     * @return scheme that was found or null
     */
    public RemoteScheme getScheme(final String name, final String type, boolean required) throws RemoteException, ClientException {

        RemoteScheme result = null;
        RemoteScheme[] list = null;
        if (!"".equals(name)) {
            if ("Issue Security".equalsIgnoreCase(type)) {
                list = service.getSecuritySchemes(token);
            } else if ("Notification".equalsIgnoreCase(type)) {
                list = service.getNotificationSchemes(token);
            } else if ("Permission".equalsIgnoreCase(type)) {
                list = service.getPermissionSchemes(token);
            }
        }

        int index = findEntityIndex(name, list);
        if (index >= 0) {
            result = list[index];
        }
        if (required && (result == null)) {
            throw new RemoteException(type + "scheme '" + name + "' not found");
        }
        return result;
    }

    /**
     * Converts a set of RemoteFieldValue's to an array
     * 
     * @param set of remote field values
     * @return array of remote field values
     */
    public RemoteFieldValue[] getRemoteFieldValueArray(final Set<RemoteFieldValue> set) {
        int i = 0;
        RemoteFieldValue[] result = new RemoteFieldValue[set.size()];
        for (RemoteFieldValue element : set) {
            result[i] = element;
            i++;
        }
        return result;
    }

    public void experiment() {

    }
}